"""
Functions/classes related to hostnames/domains/reverse DNS etc. - network related helper code

**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )

"""

import asyncio
import socket
from ipaddress import IPv4Address, IPv6Address, ip_address

from typing import AsyncGenerator, Generator, List, Optional, Tuple, Union

from privex.helpers import plugin
from privex.helpers.common import empty
from privex.helpers.exceptions import BoundaryException, InvalidHost, ReverseDNSNotFound
import logging

from privex.helpers.net.util import is_ip, sock_ver
from privex.helpers.types import IP_OR_STR

log = logging.getLogger(__name__)

__all__ = [
    'ip_to_rdns', 'ip4_to_rdns', 'ip6_to_rdns', 'resolve_ips_async', 'resolve_ip_async', 'resolve_ips_multi_async',
    'resolve_ips', 'resolve_ip', 'resolve_ips_multi', 'get_rdns_async', 'get_rdns', 'get_rdns_multi'
]


try:
    from dns.resolver import Resolver, NoAnswer, NXDOMAIN
    
    
    def asn_to_name(as_number: Union[int, str], quiet: bool = True) -> str:
        """
        Look up an integer Autonomous System Number and return the human readable
        name of the organization.

        Usage:

        >>> asn_to_name(210083)
        'PRIVEX, SE'
        >>> asn_to_name('13335')
        'CLOUDFLARENET - Cloudflare, Inc., US'

        This helper function requires ``dnspython>=1.16.0``, it will not be visible unless
        you install the dnspython package in your virtualenv, or systemwide::

            pip3 install dnspython


        :param int/str as_number: The AS number as a string or integer, e.g. 210083 or '210083'
        :param bool quiet:        (default True) If True, returns 'Unknown ASN' if a lookup fails.
                                  If False, raises a KeyError if no results are found.
        :raises KeyError:         Raised when a lookup returns no results, and ``quiet`` is set to False.
        :return str as_name:      The name and country code of the ASN, e.g. 'PRIVEX, SE'
        """
        
        try:
            res = Resolver().query('AS{}.asn.cymru.com'.format(as_number), "TXT")
            if len(res) > 0:
                # res[0] is formatted like such: "15169 | US | arin | 2000-03-30 | GOOGLE - Google LLC, US" with
                # literal quotes. we need to strip them, split by pipe, extract the last element, then strip spaces.
                asname = str(res[0]).strip('"').split('|')[-1:][0].strip()
                return str(asname)
            raise NoAnswer('privex.helpers.net.asn_to_name returned no results.')
        except (NoAnswer, NXDOMAIN):
            if quiet:
                return 'Unknown ASN'
            raise KeyError('ASN {} was not found, or server did not respond.'.format(as_number))
    
    
    __all__ += ['asn_to_name']
    plugin.HAS_DNSPYTHON = True

except ImportError:
    log.debug('privex.helpers.net failed to import "dns.resolver" (pypi package "dnspython"), skipping some helpers')
    pass


def ip_to_rdns(ip: str, boundary: bool = False, v6_boundary: int = 32, v4_boundary: int = 24) -> str:
    """
    Converts an IPv4 or IPv6 address into an in-addr domain

    Default boundaries: IPv4 - 24 bits, IPv6 - 32 bits

    **Examples:**

        >>> ip_to_rdns('127.0.0.1') # IPv4 to arpa format
            '1.0.0.127.in-addr.arpa'

        >>> ip_to_rdns('2001:dead:beef::1') # IPv6 to arpa format
            '1.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.f.e.e.b.d.a.e.d.1.0.0.2.ip6.arpa'

        >>> ip_to_rdns('2001:dead:beef::1', boundary=True) # IPv6 32-bit boundary to arpa
            'd.a.e.d.1.0.0.2.ip6.arpa'

    :param str ip: IPv4 or IPv6 address
    :param bool boundary: If True, return the base (boundary) domain to place NS/SOA
    :param int v6_boundary: Bits for IPv6 boundary. Must be dividable by 4 bits (nibble)
    :param int v4_boundary: Bits for IPv4 boundary. Must be dividable by 8 bits (octet)

    :raises ValueError: When IP address is invalid
    :raises BoundaryException: When boundary for IPv4/v6 is invalid

    :return str rdns_domain: in-addr.arpa format, e.g. ``0.0.127.in-addr.arpa``
    :return str rdns_domain: ip6.arpa format, e.g. ``0.8.e.f.ip6.arpa``
    """
    v6_boundary = int(v6_boundary)
    v4_boundary = int(v4_boundary)
    # sanity check
    if boundary:
        # will raise a BoundaryException if v4/v6 is invalid
        _check_boundaries(v4_boundary, v6_boundary)
    # ip_address helps us detect the type, and uncompress v6 addresses
    ip_obj = ip_address(ip)

    # IPv6 Addresses use nibbles and ip6.arpa
    ip_type = type(ip_obj)
    if ip_type == IPv6Address:
        return ip6_to_rdns(ip_obj, v6_boundary, boundary)
    elif ip_type == IPv4Address:
        return ip4_to_rdns(ip_obj, v4_boundary, boundary)
    else:
        raise ValueError("Not a v4 nor v6 address...? Type was: {}".format(ip_type))


def _check_boundaries(v4_boundary: int, v6_boundary: int):
    """
    Boundary validator for :py:func:`.ip_to_rdns`

    :raises BoundaryException: When either the v4 or v6 boundary is invalid
    """
    if v6_boundary > 128 or v6_boundary < 4:
        raise BoundaryException('v6_boundary must be at least 4 bits, and less than 128 (was: {})'.format(v6_boundary))
    if v4_boundary > 32 or v4_boundary < 8:
        raise BoundaryException('v4_boundary must be at least 8 bits, and less than 32 (was: {})'.format(v4_boundary))
    if v6_boundary % 4 != 0:
        raise BoundaryException('v6_boundary not dividable by 4 bits (was: {})'.format(v6_boundary))
    if v4_boundary % 8 != 0:
        raise BoundaryException('v4_boundary not dividable by 8 bits (was: {})'.format(v4_boundary))


def ip4_to_rdns(ip_obj: IPv4Address, v4_boundary: int = 24, boundary: bool = False) -> str:
    """
    Internal function for getting the rDNS domain for a given v4 address. Use :py:func:`.ip_to_rdns` unless
    you have a specific need for this one.

    :param IPv4Address ip_obj: An IPv4 ip_address() object to get the rDNS domain for
    :param int v4_boundary: 8-32 bits. If ``boundary`` is True, return the base rDNS domain at this boundary.
    :param bool boundary: If True, cut off the rDNS domain to the given ``v4_boundary``
    :return str rdns_domain: in-addr.arpa format, e.g. ``0.0.127.in-addr.arpa``
    """
    addr = ip_obj.reverse_pointer
    if boundary:
        exploded = ip_obj.exploded
        # each octet of ipv4 is 8 bits. only take the first x octets for boundary
        octets = int(v4_boundary / 8)
        addr_bounded = exploded.split('.')[:octets][::-1]  # array of octets, trim octets, and reverse the array
        addr_joined = '.'.join(addr_bounded)  # glue the array back together
        addr = addr_joined + '.in-addr.arpa'  # and finally add the arpa subdomain
    return addr


def ip6_to_rdns(ip_obj: IPv6Address, v6_boundary: int = 32, boundary: bool = False) -> str:
    """
    Internal function for getting the rDNS domain for a given v6 address. Use :py:func:`.ip_to_rdns` unless
    you have a specific need for this one.

    :param IPv6Address ip_obj: An IPv4 ip_address() object to get the rDNS domain for
    :param int v6_boundary: 8-128 bits. If ``boundary`` is True, return the base rDNS domain at this boundary.
    :param bool boundary: If True, cut off the rDNS domain to the given ``v6_boundary``
    :return str rdns_domain: ip6.arpa format, e.g. ``0.8.e.f.ip6.arpa``
    """
    # first uncompress the IP
    expanded = ip_obj.exploded
    # Remove the colons
    # if we're not doing boundary, this will be used later, otherwise it'll just be replaced anyway
    addr = expanded.replace(':', '')
    if boundary:
        # if we need a boundary address, use v6_boundary to strip off the excess characters
        del_chars = int(v6_boundary / 4)        # 4 bits = 1 character (nibble)
        addr = addr[:del_chars]                 # now we have the boundary portion, e.g. 200100fa
    addr_reverse = addr[::-1]                   # reverse the address, e.g. af001002
    addr_joined = '.'.join(list(addr_reverse))  # join together with dots, e.g. a.f.0.0.1.0.0.2
    return addr_joined + '.ip6.arpa'            # and finally, return the completed string, a.f.0.0.1.0.0.2.ip6.arpa


async def resolve_ips_async(addr: IP_OR_STR, version: Union[str, int] = 'any', v4_convert=False) -> List[str]:
    """
    AsyncIO version of :func:`.resolve_ips_async` - resolves the IPv4/v6 addresses for a given host (``addr``)
    
    
    :param str|IPv4Address|IPv6Address addr: The hostname to resolve. If an IPv4 / IPv6 address is passed instead of a hostname,
                                             it will be validated against ``version``, then returned in a single item list.

    :param str|int version: Default: ``'any'`` - Return both IPv4 and IPv6 addresses (if both are found). If an IP address is passed,
                            then both IPv4 and IPv6 addresses will be accepted. If set to one of the IPv4/IPv6 version choices, then a
                            passed IP of the wrong version will raise :class:`.AttributeError`

                            Choices:

                              * **IPv4 Options**: ``4`` (int), ``'v4'``, ``'4'`` (str), ``'ipv4'``, ``'inet'``, ``'inet4'``
                              * **IPv6 Options**: ``6`` (int), ``'v6'``, ``'6'`` (str), ``'ipv6'``, ``'inet6'``

    :param bool v4_convert: (Default: ``False``) If set to ``True``, will allow IPv6-wrapped IPv4 addresses starting with ``::ffff:`` to
                            be returned when requesting version ``v6`` from an IPv4-only hostname.

    :raises AttributeError: Raised when an IPv4 address is passed and ``version`` is set to IPv6 - as well as vice versa (IPv6 passed
                            while version is set to IPv4)

    :return List[str] ips: Zero or more IP addresses in a list of :class:`str`'s
    """
    loop = asyncio.get_event_loop()
    addr, version = str(addr), sock_ver(version)
    ips = []
    ip = is_ip(addr, version)
    if ip: return [str(ip)]
    try:
        if version in [socket.AF_INET, socket.AF_INET6]:
            _ips = await loop.getaddrinfo(addr, 2001, family=version, proto=socket.IPPROTO_TCP)
        else:
            _ips = await loop.getaddrinfo(addr, 2001, proto=socket.IPPROTO_TCP)
        
        for ip in _ips:
            ips += [ip[-1][0]]
        # If a hostname has no AAAA records, and we request AF_INET6, getaddrinfo often converts the A records
        # into IPv6-wrapped IPv4 addresses like so: ``::ffff:13.77.161.179``
        # Most people who specifically request IPv6 want only real IPv6 addresses, not IPv4 addresses wrapped in IPv6 format.
        if not v4_convert:
            ips = [ip for ip in ips if not ip.startswith('::ffff')]
        return ips
    except Exception as e:
        log.warning("Exception occurred while resolving host %s - reason: %s %s", addr, type(e), str(e))
        return ips


async def resolve_ip_async(addr: IP_OR_STR, version: Union[str, int] = 'any', v4_convert=False) -> Optional[str]:
    """
    AsyncIO version of :func:`.resolve_ip` - resolves the IPv4/v6 address for a given host (``addr``)
    """
    ips = await resolve_ips_async(addr, version=version, v4_convert=v4_convert)
    if len(ips) == 0:
        return None
    return ips[0]


async def resolve_ips_multi_async(*addr: IP_OR_STR, version: Union[str, int] = 'any', v4_convert=False) \
        -> AsyncGenerator[Tuple[str, Optional[List[str]]], None]:
    """
    Async version of :func:`.resolve_ips_multi`. Resolve IPv4/v6 addresses for multiple hosts specified as positional arguments.
    
    Returns results as an AsyncIO generator, to allow for efficient handling of a large amount of hostnames to resolve.

    Using the AsyncIO generator in a loop efficiently::

        >>> async for host, ips in resolve_ips_multi_async('privex.io', 'cloudflare.com', 'google.com'):
        ...     print(f"{host:<20} ->   {', '.join(ips)}")
        ...
        privex.io            ->   2a07:e00::abc, 185.130.44.10
        cloudflare.com       ->   2606:4700::6811:af55, 2606:4700::6811:b055, 104.17.176.85, 104.17.175.85
        google.com           ->   2a00:1450:4009:807::200e, 216.58.204.238

    If you're only resolving a small number of hosts ( less than 100 or so ), you can simply comprehend the generator into
    a :class:`list` and then turn the result into a :class:`dict` using ``dict()``, which will get you a dictionary of hosts mapped to
    lists of IP addresses.

    Dictionary Cast Examples::

        >>> ips = [x async for x in resolve_ips_multi_async('privex.io', 'microsoft.com', 'google.com')]
        >>> dict(ips)
        {'privex.io': ['2a07:e00::abc', '185.130.44.10'],
         'microsoft.com': ['104.215.148.63', '40.76.4.15', '40.112.72.205', '40.113.200.201', '13.77.161.179'],
         'google.com': ['2a00:1450:4009:807::200e', '216.58.204.238']}
        >>> dict([x async for x in resolve_ips_multi_async('privex.io', 'microsoft.com', 'google.com', version='v6')])
        {'privex.io': ['2a07:e00::abc'], 'microsoft.com': [], 'google.com': ['2a00:1450:4009:81c::200e']}
        >>> dict([x async for x in resolve_ips_multi_async('privex.io', 'this-does-not-exist', 'google.com', version='v6')])
        {'privex.io': ['2a07:e00::abc'], 'this-does-not-exist': [], 'google.com': ['2a00:1450:4009:81c::200e']}
         >>> dict([x async for x in resolve_ips_multi_async('privex.io', 'example.com', '127.0.0.1', version='v6')])
        [resolve_ips_multi AttributeError] Invalid IP: 127.0.0.1 - Ex: <class 'AttributeError'> Passed address '127.0.0.1' was
                                           an IPv4 address, but 'version' requested an IPv6 address.
        {'privex.io': ['2a07:e00::abc'], 'example.com': ['2606:2800:220:1:248:1893:25c8:1946'], '127.0.0.1': None}


    :param str|IPv4Address|IPv6Address addr: Hostname's to resolve / IP addresses to validate or pass-thru
    :param str|int version: (Default: ``any``) - ``4`` (int), ``'v4'``, ``6`` (int), ``'v6'`` (see :func:`.resolve_ips` for more options)
    :param bool v4_convert: (Default: ``False``) If set to ``True``, will allow IPv6-wrapped IPv4 addresses starting with ``::ffff:`` to
                            be returned when requesting version ``v6`` from an IPv4-only hostname.

    :return Tuple[str,Optional[List[str]] gen:  An async generator which returns tuples containing a hostname/IP, and a list of it's
        resolved IPs. If the IP was rejected (e.g. IPv4 IP passed with ``v6`` ``version`` param), then the list may instead be ``None``.
    
    """
    for a in addr:
        try:
            res = await resolve_ips_async(a, version=version, v4_convert=v4_convert)
            yield (a, res)
        except socket.gaierror as e:
            log.warning("[resolve_ips_multi socket.gaierror] Failed to resolve host: %s - Ex: %s %s", addr, type(e), str(e))
            yield (a, None)
        except AttributeError as e:
            log.warning("[resolve_ips_multi AttributeError] Invalid IP: %s - Ex: %s %s", a, type(e), str(e))
            yield (a, None)


def resolve_ips(addr: IP_OR_STR, version: Union[str, int] = 'any', v4_convert=False) -> List[str]:
    """
    With just a single hostname argument, both IPv4 and IPv6 addresses will be returned as strings::
    
        >>> resolve_ips('www.privex.io')
        ['2a07:e00::abc', '185.130.44.10']
    
    You can provide the ``version`` argument as either positional or kwarg, e.g. ``v4`` or ``v6`` to restrict the results
    to only that IP version::
    
        >>> resolve_ips('privex.io', version='v4')
        ['185.130.44.10']
    
    The ``v4_convert`` option is ``False`` by default, which prevents ``::ffff:`` style IPv6 wrapped IPv4 addresses being
    returned when you request version ``v6``::
    
        >>> resolve_ips('microsoft.com')
        ['40.76.4.15', '40.112.72.205', '13.77.161.179', '40.113.200.201', '104.215.148.63']
        >>> resolve_ips('microsoft.com', 'v6')
        []
    
    If for whatever reason, you need ``::ffff:`` IPv6 wrapped IPv4 addresses to be returned, you can set ``v4_convert=True``,
    which will disable filtering out ``::ffff:`` fake IPv6 addresses::
    
        >>> resolve_ips('microsoft.com', 'v6', v4_convert=True)
        ['::ffff:40.76.4.15', '::ffff:40.112.72.205', '::ffff:13.77.161.179',
         '::ffff:40.113.200.201', '::ffff:104.215.148.63']
    
    For convenience, if an IPv4 / IPv6 address is specified, then it will simply be validated against ``version`` and then returned
    within a list. This is useful when handling user specified data, which may be either a hostname or an IP address, and you
    need to resolve hostnames while leaving IP addresses alone::
    
        >>> resolve_ips('8.8.4.4')
        ['8.8.4.4']
        >>> resolve_ips('2a07:e00::333')
        ['2a07:e00::333']
        
        >>> resolve_ips('8.8.4.4', 'v6')
        Traceback (most recent call last):
          File "<ipython-input-10-6ca9e766006f>", line 1, in <module>
            resolve_ips('8.8.4.4', 'v6')
        AttributeError: Passed address '8.8.4.4' was an IPv4 address, but 'version' requested an IPv6 address.
        
        >>> resolve_ips('2a07:e00::333', 'v4')
        Traceback (most recent call last):
          File "<ipython-input-11-543bfa71c57a>", line 1, in <module>
            resolve_ips('2a07:e00::333', 'v4')
        AttributeError: Passed address '2a07:e00::333' was an IPv6 address, but 'version' requested an IPv4 address.
        

    :param str|IPv4Address|IPv6Address addr: The hostname to resolve. If an IPv4 / IPv6 address is passed instead of a hostname,
                                             it will be validated against ``version``, then returned in a single item list.
    
    :param str|int version: Default: ``'any'`` - Return both IPv4 and IPv6 addresses (if both are found). If an IP address is passed,
                            then both IPv4 and IPv6 addresses will be accepted. If set to one of the IPv4/IPv6 version choices, then a
                            passed IP of the wrong version will raise :class:`.AttributeError`
                            
                            Choices:
                            
                              * **IPv4 Options**: ``4`` (int), ``'v4'``, ``'4'`` (str), ``'ipv4'``, ``'inet'``, ``'inet4'``
                              * **IPv6 Options**: ``6`` (int), ``'v6'``, ``'6'`` (str), ``'ipv6'``, ``'inet6'``
    
    :param bool v4_convert: (Default: ``False``) If set to ``True``, will allow IPv6-wrapped IPv4 addresses starting with ``::ffff:`` to
                            be returned when requesting version ``v6`` from an IPv4-only hostname.
    
    :raises AttributeError: Raised when an IPv4 address is passed and ``version`` is set to IPv6 - as well as vice versa (IPv6 passed
                            while version is set to IPv4)
    
    :return List[str] ips: Zero or more IP addresses in a list of :class:`str`'s
    """
    addr, version, ips = str(addr), sock_ver(version), []
    ip = is_ip(addr, version)
    if ip: return [str(ip)]
    try:
        if version in [socket.AF_INET, socket.AF_INET6]:
            _ips = socket.getaddrinfo(addr, 2001, family=version, proto=socket.IPPROTO_TCP)
        else:
            _ips = socket.getaddrinfo(addr, 2001, proto=socket.IPPROTO_TCP)
        for ip in _ips:
            ips += [ip[-1][0]]
        # If a hostname has no AAAA records, and we request AF_INET6, getaddrinfo often converts the A records
        # into IPv6-wrapped IPv4 addresses like so: ``::ffff:13.77.161.179``
        # Most people who specifically request IPv6 want only real IPv6 addresses, not IPv4 addresses wrapped in IPv6 format.
        if not v4_convert:
            ips = [ip for ip in ips if not ip.startswith('::ffff')]
        return ips
    except Exception as e:
        log.warning("Exception occurred while resolving host %s - reason: %s %s", addr, type(e), str(e))
        return ips


def resolve_ip(addr: IP_OR_STR, version: Union[str, int] = 'any', v4_convert=False) -> Optional[str]:
    """
    Wrapper for :func:`.resolve_ips` - passes args to :func:`.resolve_ips` and returns the first item from the results.
    
    If the results are empty, ``None`` will be returned.
    
    Examples::
    
        >>> resolve_ip('privex.io')
        '2a07:e00::abc'
        >>> resolve_ip('privex.io', 'v4')
        '185.130.44.10'
        >>> resolve_ip('microsoft.com')
        '104.215.148.63'
        >>> repr(resolve_ip('microsoft.com' ,'v6'))
        'None'
        >>> resolve_ip('microsoft.com' ,'v6', v4_convert=True)
        '::ffff:104.215.148.63'

    :param str|IPv4Address|IPv6Address addr: Hostname to resolve / IP address to validate or pass-thru
    
    :param str|int version: (Default: ``any``) - ``4`` (int), ``'v4'``, ``6`` (int), ``'v6'`` (see :func:`.resolve_ips` for more options)
    
    :param bool v4_convert: (Default: ``False``) If set to ``True``, will allow IPv6-wrapped IPv4 addresses starting with ``::ffff:`` to
                            be returned when requesting version ``v6`` from an IPv4-only hostname.
    
    :raises AttributeError: Raised when an IPv4 address is passed and ``version`` is set to IPv6 - as well as vice versa (IPv6 passed
                            while version is set to IPv4)
    
    :return Optional[str] ips: An IPv4/v6 address as a string if there was at least 1 result - otherwise ``None``.
    """
    ips = resolve_ips(addr, version=version, v4_convert=v4_convert)
    if len(ips) == 0:
        return None
    return ips[0]


def resolve_ips_multi(*addr: IP_OR_STR, version: Union[str, int] = 'any', v4_convert=False) \
        -> Generator[Tuple[str, Optional[List[str]]], None, None]:
    """
    Resolve IPv4/v6 addresses for multiple hosts specified as positional arguments.
    
    Returns results as a generator, to allow for efficient handling of a large amount of hostnames to resolve.
    
    Using the generator in a loop efficiently::
    
        >>> for host, ips in resolve_ips_multi('privex.io', 'cloudflare.com', 'google.com'):
        ...     print(f"{host:<20} ->   {', '.join(ips)}")
        ...
        privex.io            ->   2a07:e00::abc, 185.130.44.10
        cloudflare.com       ->   2606:4700::6811:af55, 2606:4700::6811:b055, 104.17.176.85, 104.17.175.85
        google.com           ->   2a00:1450:4009:807::200e, 216.58.204.238
    
    If you're only resolving a small number of hosts ( less than 100 or so ), you can simply cast the generator into
    a :class:`dict` using ``dict()``, which will get you a dictionary of hosts mapped to lists of IP addresses.

    Dictionary Cast Examples::
        
        >>> dict(resolve_ips_multi('privex.io', 'microsoft.com', 'google.com'))
        {'privex.io': ['2a07:e00::abc', '185.130.44.10'],
         'microsoft.com': ['104.215.148.63', '40.76.4.15', '40.112.72.205', '40.113.200.201', '13.77.161.179'],
         'google.com': ['2a00:1450:4009:807::200e', '216.58.204.238']}
        >>> dict(resolve_ips_multi('privex.io', 'microsoft.com', 'google.com', version='v6'))
        {'privex.io': ['2a07:e00::abc'], 'microsoft.com': [], 'google.com': ['2a00:1450:4009:81c::200e']}
        >>> dict(resolve_ips_multi('privex.io', 'this-does-not-exist', 'google.com', version='v6'))
        {'privex.io': ['2a07:e00::abc'], 'this-does-not-exist': [], 'google.com': ['2a00:1450:4009:81c::200e']}
         >>> dict(resolve_ips_multi('privex.io', 'example.com', '127.0.0.1', version='v6'))
        [resolve_ips_multi AttributeError] Invalid IP: 127.0.0.1 - Ex: <class 'AttributeError'> Passed address '127.0.0.1' was
                                           an IPv4 address, but 'version' requested an IPv6 address.
        {'privex.io': ['2a07:e00::abc'], 'example.com': ['2606:2800:220:1:248:1893:25c8:1946'], '127.0.0.1': None}
    
    
    :param str|IPv4Address|IPv6Address addr: Hostname to resolve / IP address to validate or pass-thru
    :param str|int version: (Default: ``any``) - ``4`` (int), ``'v4'``, ``6`` (int), ``'v6'`` (see :func:`.resolve_ips` for more options)
    :param bool v4_convert: (Default: ``False``) If set to ``True``, will allow IPv6-wrapped IPv4 addresses starting with ``::ffff:`` to
                            be returned when requesting version ``v6`` from an IPv4-only hostname.
                            
    :return Tuple[str,Optional[List[str]] gen:  A generator which returns tuples containing a hostname/IP, and a list of it's resolved
        IPs. If the IP was rejected (e.g. IPv4 IP passed with ``v6`` ``version`` param), then the list may instead be ``None``.
    """
    for a in addr:
        try:
            res = resolve_ips(a, version=version, v4_convert=v4_convert)
            yield (a, res)
        except socket.gaierror as e:
            log.warning("[resolve_ips_multi socket.gaierror] Failed to resolve host: %s - Ex: %s %s", addr, type(e), str(e))
            yield (a, None)
        except AttributeError as e:
            log.warning("[resolve_ips_multi AttributeError] Invalid IP: %s - Ex: %s %s", a, type(e), str(e))
            yield (a, None)


async def get_rdns_async(host: IP_OR_STR, throw=True, version='any', name_port=80) -> Optional[str]:
    """
    AsyncIO version of :func:`.get_rdns` - get the reverse DNS for a given host (IP address or domain)
    
        >>> from privex.helpers import get_rdns_async
        >>> await get_rdns_async('185.130.44.10')
        'web-se1.privex.io'
        >>> await get_rdns_async('2a07:e00::333')
        'se.dns.privex.io'
        >>> await get_rdns_async('privex.io')
        'web-se1.privex.io'
    
    :param str|IPv4Address|IPv6Address host: An IPv4/v6 address, or domain to lookup reverse DNS for.
    :param bool throw: (Default: ``True``) When ``True``, will raise :class:`.ReverseDNSNotFound` or :class:`.InvalidHost` when no
                       rDNS records can be found for ``host``, or when ``host`` is an invalid IP / non-existent domain.
                       When ``False``, will simply return ``None`` when ``host`` is invalid, or no rDNS records are found.
    
    :param str|int version: IP version to use when looking up a domain/hostname (default: ``'any'``)
    
    :param int name_port: This generally isn't important. This port is passed to :func:`loop.getnameinfo` when looking
                          up the reverse DNS for ``host``. Usually there's no reason to change this from the default.
    
    :raises ReverseDNSNotFound: When ``throw`` is True and no rDNS records were found for ``host``
    :raises InvalidHost: When ``throw`` is True and ``host`` is an invalid IP address or non-existent domain/hostname
    :return Optional[str] rDNS: The reverse DNS hostname for ``host`` (value of PTR record)
    """
    loop = asyncio.get_event_loop()
    host = str(host)
    try:
        if not is_ip(host):
            orig_host = host
            host = await resolve_ip_async(host, version=version)
            if empty(host):
                if throw:
                    raise InvalidHost(f"Host '{orig_host}' is not a valid IP address, nor an existent domain")
                return None
        
        res = await loop.getnameinfo((host, name_port))
        rdns = res[0]
        if is_ip(rdns):
            if throw: raise ReverseDNSNotFound(f"No reverse DNS records found for host '{host}' - result was: {rdns}")
            return None
        return rdns
    except socket.gaierror as e:
        if throw: raise InvalidHost(f"Host '{host}' is not a valid IP address, nor an existent domain: {type(e)} {str(e)}")
    return None


def get_rdns(host: IP_OR_STR, throw=True) -> Optional[str]:
    """
    Look up the reverse DNS hostname for ``host`` and return it as a string. The ``host`` can be an IP address as a :class:`str`,
    :class:`.IPv4Address`, :class:`.IPv6Address` - or a domain.
    
    If a domain is passed, e.g. ``privex.io`` - then the reverse DNS will be looked up for the IP address contained in the domain's
    AAAA or A records.
    
    Toggle ``throw`` to control whether to raise exceptions on error (``True``), or to simply return None (``False``).
    
    Basic usage::
    
        >>> from privex.helpers import get_rdns
        >>> get_rdns('185.130.44.10')
        'web-se1.privex.io'
        >>> get_rdns('2a07:e00::333')
        'se.dns.privex.io'
        >>> get_rdns('privex.io')
        'web-se1.privex.io'
    
    Error handling::
    
        >>> get_rdns('192.168.4.5')
        Traceback (most recent call last):
          File "<ipython-input-14-e1ed65295031>", line 1, in <module>
            get_rdns('192.168.4.5')
        privex.helpers.exceptions.ReverseDNSNotFound: No reverse DNS records found for host '192.168.4.5':
            <class 'socket.herror'> [Errno 1] Unknown host
        >>> get_rdns('non-existent-domain.example')
        Traceback (most recent call last):
          File "<ipython-input-16-0d75d37a930f>", line 1, in <module>
            get_rdns('non-existent-domain.example')
        privex.helpers.exceptions.InvalidHost: Host 'non-existent-domain.example' is not a valid IP address,
        nor an existent domain: <class 'socket.gaierror'> [Errno 8] nodename nor servname provided, or not known
        >>> repr(get_rdns('192.168.4.5', throw=False))
        'None'
        >>> repr(get_rdns('non-existent-domain.example', False))
        'None'
    
    
    :param str|IPv4Address|IPv6Address host: An IPv4/v6 address, or domain to lookup reverse DNS for.
    :param bool throw: (Default: ``True``) When ``True``, will raise :class:`.ReverseDNSNotFound` or :class:`.InvalidHost` when no
                       rDNS records can be found for ``host``, or when ``host`` is an invalid IP / non-existent domain.
                       When ``False``, will simply return ``None`` when ``host`` is invalid, or no rDNS records are found.
    :raises ReverseDNSNotFound: When ``throw`` is True and no rDNS records were found for ``host``
    :raises InvalidHost: When ``throw`` is True and ``host`` is an invalid IP address or non-existent domain/hostname
    :return Optional[str] rDNS: The reverse DNS hostname for ``host`` (value of PTR record)
    """
    try:
        rdns = socket.gethostbyaddr(str(host))
        return rdns[0]
    except socket.herror as e:
        if throw: raise ReverseDNSNotFound(f"No reverse DNS records found for host '{host}': {type(e)} {str(e)}")
    except socket.gaierror as e:
        if throw: raise InvalidHost(f"Host '{host}' is not a valid IP address, nor an existent domain: {type(e)} {str(e)}")
    return None


def get_rdns_multi(*hosts: IP_OR_STR, throw=False) -> Generator[Tuple[str, Optional[str]], None, None]:
    """
    Resolve reverse DNS hostnames for multiple IPs / domains specified as positional arguments.
    
    Each host in ``hosts`` can be an IP address as a :class:`str`, :class:`.IPv4Address`, :class:`.IPv6Address` - or a domain.
    
    Returns results as a generator, to allow for efficient handling of a large amount of hosts to resolve.
    
    Basic usage::
    
        >>> for host, rdns in get_rdns_multi('185.130.44.10', '8.8.4.4', '1.1.1.1', '2a07:e00::333'):
        >>>     print(f"{host:<20} -> {rdns:>5}")
        185.130.44.10        -> web-se1.privex.io
        8.8.4.4              -> dns.google
        1.1.1.1              -> one.one.one.one
        2a07:e00::333        -> se.dns.privex.io
    
    If you're only resolving a small number of hosts ( less than 100 or so ), you can simply cast the generator into
    a :class:`dict` using ``dict()``, which will get you a dictionary of hosts mapped to their rDNS::
    
        >>> data = dict(get_rdns_multi('185.130.44.10', '8.8.4.4', '1.1.1.1', '2a07:e00::333'))
        >>> data['8.8.4.4']
        'dns.google'
        >>> data.get('2a07:e00::333', 'error')
        'se.dns.privex.io'
    
    :param str|IPv4Address|IPv6Address hosts: One or more IPv4/v6 addresses, or domains to lookup reverse DNS for - as positional args.
    
    :param bool throw: (Default: ``False``) When ``True``, will raise :class:`.ReverseDNSNotFound` or :class:`.InvalidHost` when no
                       rDNS records can be found for a host, or when the host is an invalid IP / non-existent domain.
                       When ``False``, will simply return ``None`` when a host is invalid, or no rDNS records are found.
    
    :raises ReverseDNSNotFound: When ``throw`` is True and no rDNS records were found for ``host``
    :raises InvalidHost: When ``throw`` is True and ``host`` is an invalid IP address or non-existent domain/hostname
    
    :return Tuple[str,Optional[str]] rDNS: A generator returning :class:`tuple`'s containing the original passed host, and it's
                                           reverse DNS hostname (value of PTR record)
    """
    for h in hosts:
        yield (str(h), get_rdns(str(h), throw=throw))



"""
Helper functions/classes related to caching.

This module acts as a singleton wrapper, allowing for easily setting a framework-independent global cache API.

To make the module easy to use, :py:func:`.adapter_get` initialises an instance of :class:`.MemoryCache` if no
global cache adapter instance has been setup. This means you can use the various alias functions in this module
without having to configure a cache adapter.

Available Cache Adapters
^^^^^^^^^^^^^^^^^^^^^^^^

**Standard Synchronous Adapters**

Four synchronous cache adapters are included by default - :class:`.MemoryCache` (dependency free),
:class:`.MemcachedCache` (needs ``pylibmc`` library), :class:`.SqliteCache` (needs ``privex-db`` library),
:class:`.RedisCache` (needs ``redis`` library).

While these synchronous classes don't support coroutines for most methods, as of privex-helpers 2.7 the method
:meth:`privex.helpers.cache.CacheAdapter.CacheAdapter.get_or_set_async` is an async version of :meth:`.CacheAdapter.get_or_set`,
and is available on all :class:`.CacheAdapter` sub-classes (both :class:`.MemoryCache` and :class:`.RedisCache`).
``get_or_set_async`` allows a coroutine or coroutine function/method reference to be passed as the fallback value.

    ==============================   ==================================================================================================
    Adapter                          Description
    ==============================   ==================================================================================================
    :class:`.CacheAdapter`           This is the base class for all synchronous cache adapters (doesn't do anything)
    :class:`.MemoryCache`            A cache adapter which stores cached items in memory using a dict. Fully functional incl. timeout.
    :class:`.MemcachedCache`         A cache adapter for `Memcached`_ using the synchronous python library ``pylibmc``
    :class:`.RedisCache`             A cache adapter for `Redis`_ using the python library ``redis``
    :class:`.SqliteCache`            A cache adapter for `SQLite3`_ using the standard Python module :mod:`sqlite3` + :mod:`privex.db`
    ==============================   ==================================================================================================

**Asynchronous (Python AsyncIO) Adapters**

Over the past few years, Python's AsyncIO has grown more mature and has gotten a lot of attention. Thankfully, whether you use
AsyncIO or not, we've got you covered.

Four AsyncIO cache adapters are included by default - :class:`.AsyncMemoryCache` (dependency free),
:class:`.AsyncRedisCache` (needs ``aioredis`` library), :class:`.AsyncSqliteCache` (needs ``aiosqlite`` library),
and :class:`.AsyncMemcachedCache` (needs ``aiomcache`` library).

    ==============================   ==================================================================================================
    Adapter                          Description
    ==============================   ==================================================================================================
    :class:`.AsyncCacheAdapter`      This is the base class for all AsyncIO cache adapters (abstract class, only implements get_or_set)
    :class:`.AsyncMemoryCache`       A cache adapter which stores cached items in memory using a dict. Fully functional incl. timeout.
    :class:`.AsyncRedisCache`        A cache adapter for `Redis`_ using the AsyncIO python library ``aioredis``
    :class:`.AsyncMemcachedCache`    A cache adapter for `Memcached`_ using the AsyncIO python library ``aiomcache``
    :class:`.AsyncSqliteCache`       A cache adapter for `SQLite3`_ using the AsyncIO python library ``aiosqlite``
    ==============================   ==================================================================================================


.. _Redis: https://redis.io/
.. _Memcached: https://www.memcached.org/
.. _SQLite3: https://www.sqlite.org/

Setting / updating the global cache adapter instance
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

First import the ``cache`` module.

    >>> from privex.helpers import cache

When setting an adapter using :func:`.adapter_set`, if your application has a user configurable cache adapter via
a plain text configuration file (e.g. a ``.env`` file), or you simply don't have any need to manually instantiate a cache adapter,
then you can pass either an alias name (``memory``, ``redis``, ``memcached``, ``sqlite3``), or a full adapter class name
(such as ``MemoryCache``, ``MemcachedCache``, ``RedisCache``, ``SqliteCache``). Example usage::

    >>> cache.adapter_set('memcached')
    >>> cache.adapter_set('MemcachedCache')

Alternatively, you may instantiate your cache adapter of choice before passing it to :py:func:`.adapter_set` - which updates
the global cache adapter instance.

    >>> my_adapter = cache.MemoryCache()
    >>> cache.adapter_set(my_adapter)

Once you've set the adapter, you can use the module functions such as :py:func:`.get` and :py:func:`.set` - or you
can import ``cached`` to enable dictionary-like cache item access.

    >>> cache.set('hello', 'world')
    >>> cache.get('hello')
    'world'
    >>> from privex.helpers import cached
    >>> cached['hello']
    'world'
    >>> cached['otherkey'] = 'testing'

You can also use AsyncIO adapters with the global cache adapter wrapper. :class:`.CacheWrapper` uses :func:`.awaitable` to
ensure that AsyncIO adapters can work synchronously when being called from a synchronous function, while working asynchronously
from a non-async function.


    >>> my_adapter = cache.AsyncRedisCache()
    >>> cache.adapter_set(my_adapter)
    >>>
    >>> # get_hello_async() is async, so @awaitable returns the normal .get() coroutine for awaiting
    >>> async def get_hello_async():
    ...     result = await cached.get('hello')
    ...     return result
    ...
    >>> # get_hello() is synchronous, so @awaitable seamlessly runs .get() in an event loop and returns
    >>> # the result - get_hello() can treat it as if it were just a normal synchronous function.
    >>> def get_hello():
    ...     return cached.get('hello')
    ...
    >>> get_hello()
    'world'
    >>> await get_hello_async()
    'world'

Important info about using the cache abstraction layer with AsyncIO
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

While ``cached`` **usually works** in AsyncIO contexts, due to the synchronous backwards compatibility wrappers, the standard
:class:`.CacheWrapper` can behave strangely in complex AsyncIO applications.

If your application / library is **primarily AsyncIO**, you should use :attr:`.async_cached` ( :class:`.AsyncCacheWrapper` )
instead of the standard :attr:`.cached` - and similarly, the adapter management functions :func:`.async_adapter_get` and
:func:`.async_adapter_set`.

The :class:`.AsyncCacheWrapper` adapter wrapper class - as the name implies - is generally **only** usable from async functions/methods,
and maintains a separate cache adapter instance than :class:`.CacheWrapper` ( :attr:`.cached` ), which **must be an AsyncIO adapter**
such as :class:`.AsyncRedisCache`

The AsyncIO cache abstraction layer / wrapper works pretty much exactly the same as the "hybrid" :class:`.CacheWrapper`,
but to make it clear how it can be used, here's some example code which shows setting/getting the global Async adapter, and
using the cache abstraction layer.

**Examples**

First we'll import the three main AsyncIO cache abstraction functions, plus :class:`.AsyncRedisCache` ::

    >>> from privex.helpers.cache import async_cached, async_adapter_get, async_adapter_set, AsyncRedisCache

We can use the global AsyncIO cache adapter instance (defaults to :class:`.AsyncMemoryCache`) via ``async_cached``::

    >>> await async_cached.set('hello', 'world')
    >>> await async_cached.get('hello')
    'world'

Much like the standard :class:`.CacheWrapper`, you can get and set keys using dict-like syntax, however
since ``__getitem__`` and ``__setitem__`` can't be natively async without Python complaining, they use the wrapper decorator
:func:`.awaitable` for getting, and the synchronous async wrapper function :func:`.loop_run` for setting. The use of these wrappers
may cause problems in certain scenarios, so it's recommended to avoid using the dict-like cache syntax within AsyncIO code::
 
    >>> await async_cached['hello']
    'world'
    >>> async_cached['lorem'] = 'ipsum'

To set / replace the global AsyncIO cache adapter, use :func:`.async_adapter_set` - similarly, you can use :func:`.async_adapter_get`
to get the current adapter instance (e.g. a direct instance of :class:`.AsyncRedisCache` if that's the current adapter)::

    >>> async_adapter_set(AsyncRedisCache())    # Either pass an instance of an async cache adapter class
    >>> async_adapter_set('redis')              # Or pass a simple string alias name, such as: redis, memcached, memory, sqlite3
    >>> adp = async_adapter_get()
    >>> await adp.set('lorem', 'test')  # Set 'lorem' using the AsyncRedisCache instance directly
    >>> await adp.get('lorem')          # Get 'lorem' using the AsyncRedisCache instance directly
    'test'
    >>> await async_cached.get('lorem') # Get 'lorem' using the global AsyncIO cache wrapper
    'test'




Plug-n-play usage
^^^^^^^^^^^^^^^^^

As explained near the start of this module's documentation, you don't have to set the global adapter if you only
plan on using the simple :class:`.MemoryCache` adapter.

Just start using the global cache API via either :py:mod:`privex.helpers.cache` or :py:mod:`privex.helpers.cache.cached`
and MemoryCache will automatically be instantiated as the global adapter as soon as something attempts to access
the global instance.

We recommend importing ``cached`` rather than ``cache``, as it acts as a wrapper that allows dictionary-like
cache key getting/setting, and is also immediately aware when the global cache adapter is set/replaced.

    >>> from privex.helpers import cached

You can access ``cached`` like a dictionary to get and set cache keys (they will use the default expiry time of
:py:attr:`privex.helpers.settings.DEFAULT_CACHE_TIMEOUT`)

    >>> cached['testing'] = 123
    >>> cached['testing']
    123


You can also call methods such as :py:func:`.get` and :py:func:`.set` for getting/setting cache items with more
control, for example:

1. Setting a custom expiration, or disabling expiration by setting timeout to ``None``
  
    >>> cached.set('example', 'test', timeout=30)   # Drop 'example' from the cache after 30 seconds from now.
    >>> cached.set('this key', 'is forever!', timeout=None) # A timeout of ``None`` disables automatic expiration.

2. Fallback values when a key isn't found, or have it throw an exception if it's not found instead.

    >>> cached.get('example', 'NOT FOUND')          # If the key 'example' doesn't exist, return 'NOT FOUND'
    'test'
    
    >>> try:   # By setting ``fail`` to True, ``get`` raises ``CacheNotFound`` if the key doesn't exist / is expired
    ...     cached.get('nonexistent', fail=True)
    ... except CacheNotFound:
    ...     log.error('The cache key "nonexistent" does not exist!')
    >>>

3. Using :py:func:`.get_or_set` you can specify either a standard type (e.g. ``str``, ``int``, ``dict``), or even
   a custom function to call to obtain the value to set and return.
 
    >>> cached.get_or_set('hello', lambda key: 'world', timeout=60)
    >>> cached['hello']
    'world'


**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2020     Privex Inc.   ( https://www.privex.io )


Cache API Docs
^^^^^^^^^^^^^^

(the above heading is for sidebar display purposes on the docs)


"""
import asyncio
import logging
import importlib
from inspect import isclass

from privex.helpers import plugin, settings
from privex.helpers.common import empty_if, LayeredContext
from privex.helpers.asyncx import awaitable, await_if_needed, loop_run

from privex.helpers.collections import DictObject

log = logging.getLogger(__name__)

from typing import Any, Optional, Union, Type, List
from privex.helpers.cache.CacheAdapter import CacheAdapter
from privex.helpers.cache.MemoryCache import MemoryCache


if plugin.HAS_PRIVEX_DB in [True, None]:
    try:
        from privex.helpers.cache.SqliteCache import SqliteCache
        plugin.HAS_PRIVEX_DB = True
    except ImportError:
        plugin.HAS_PRIVEX_DB = False
        log.debug(
            "[%s] Failed to import %s from %s (missing package 'privex-db' maybe?)", __name__, 'SqliteCache', f'{__name__}.SqliteCache'
        )
else:
    log.debug("[%s] Not attempting to import %s from %s as plugin check var '%s' is False.",
              __name__, 'SqliteCache', f'{__name__}.SqliteCache', 'HAS_PRIVEX_DB')

try:
    from privex.helpers.cache.RedisCache import RedisCache
except ImportError:
    log.debug("[%s] Failed to import %s from %s (missing package 'redis' maybe?)", __name__, 'RedisCache', f'{__name__}.RedisCache')

try:
    from privex.helpers.cache.MemcachedCache import MemcachedCache
except ImportError:
    log.debug("[%s] Failed to import %s from %s (missing package 'pylibmc' maybe?)", __name__, 'MemcachedCache',
              f'{__name__}.MemcachedCache')

try:
    from privex.helpers.cache.asyncx import *
except ImportError:
    log.exception("[%s] Failed to import %s from %s (unknown error!)", __name__, '*', f'{__name__}.asyncx')

from privex.helpers.exceptions import NotConfigured, CacheNotFound
from privex.helpers.settings import DEFAULT_CACHE_TIMEOUT


__STORE = {}

__STORE['adapter']: CacheAdapter
__STORE['async_adapter']: AsyncCacheAdapter


CLSCacheAdapter = Union[Type[CacheAdapter], Type[AsyncCacheAdapter]]
INSCacheAdapter = Union[CacheAdapter, AsyncCacheAdapter]
ANYCacheAdapter = Union[CLSCacheAdapter, INSCacheAdapter, str]

ADAPTER_MAP = DictObject(
    sync=DictObject(
        memory='privex.helpers.cache.MemoryCache.MemoryCache',
        redis='privex.helpers.cache.RedisCache.RedisCache',
        memcached='privex.helpers.cache.MemcachedCache.MemcachedCache',
        sqlite3='privex.helpers.cache.SqliteCache.SqliteCache',
    ),
    asyncio=DictObject(
        memory='privex.helpers.cache.asyncx.AsyncMemoryCache.AsyncMemoryCache',
        redis='privex.helpers.cache.asyncx.AsyncRedisCache.AsyncRedisCache',
        memcached='privex.helpers.cache.asyncx.AsyncMemcachedCache.AsyncMemcachedCache',
        sqlite3='privex.helpers.cache.asyncx.AsyncSqliteCache.AsyncSqliteCache',
    ),
    
)

_AM = ADAPTER_MAP

_AM.sync.ram, _AM.asyncio.ram = _AM.sync.mem, _AM.asyncio.mem = _AM.sync.memory, _AM.asyncio.memory
_AM.sync.mcache, _AM.asyncio.mcache = _AM.sync.memcache, _AM.asyncio.memcache = _AM.sync.memcached, _AM.asyncio.memcached
_AM.sync.sqlitedb, _AM.asyncio.sqlitedb = _AM.sync.sqlite, _AM.asyncio.sqlite = _AM.sync.sqlite3, _AM.asyncio.sqlite3

_AM.shared = DictObject(
    # Synchronous cache adapters
    MemoryCache=_AM.sync.memory, RedisCache=_AM.sync.redis, MemcachedCache=_AM.sync.memcached,
    SqliteCache=_AM.sync.sqlite3,
    # Synchronous cache adapters (xxxAdapter aliases)
    MemoryAdapter=_AM.sync.memory, RedisAdapter=_AM.sync.redis, MemcachedAdapter=_AM.sync.memcached,
    SqliteAdapter=_AM.sync.sqlite3,
    # AsyncIO cache adapters
    AsyncMemoryCache=_AM.asyncio.memory, AsyncRedisCache=_AM.asyncio.redis, AsyncMemcachedCache=_AM.asyncio.memcached,
    AsyncSqliteCache=_AM.asyncio.sqlite3,
    # AsyncIO cache adapters (xxxAdapter aliases)
    AsyncMemoryAdapter=_AM.asyncio.memory, AsyncRedisAdapter=_AM.asyncio.redis, AsyncMemcachedAdapter=_AM.asyncio.memcached,
    AsyncSqliteAdapter=_AM.asyncio.sqlite3
)


def import_adapter_path(full_path: str) -> Type[Union[CacheAdapter, AsyncCacheAdapter]]:
    """
    This function imports a fully qualified module path, with a class/function/attribute name as the final component
    of the dot delimitered string.
    
    Example::
    
        >>> # Equivalent to: from privex.helpers.cache.MemcachedCache import MemcachedCache as _MemcachedCache
        >>> _MemcachedCache = import_adapter_path('privex.helpers.cache.MemcachedCache.MemcachedCache')
        
    :param str full_path: A fully qualified module path, with a class/function/attribute name as the final component
                          of the dot delimitered string.
    :return type obj:     The class/function/variable that was imported from the module in ``full_path``
    """
    
    split_mod = full_path.split('.')
    path_mod, path_class = '.'.join(split_mod[:-1]), split_mod[-1]
    mod = importlib.import_module(path_mod)
    return getattr(mod, path_class)


def import_adapter(name: str, cat: str = 'sync', fallback_shared=True) -> Type[Union[CacheAdapter, AsyncCacheAdapter]]:
    """
    Import a cache adapter class using either it's class name, or config/human-friendly alias name.
    
    By default, ``fallback_shared`` is enabled, which allows importing adapter names that are in the ``shared`` category, such as
    ``AsyncRedisCache`` / ``MemcachedCache`` while ``cat`` is set to ``asyncio`` / ``sync``
    
    Adapter Categories::
    
      * ``shared`` - Contains full cache adapter class names, e.g. ``RedisCache``, ``AsyncMemcachedCache`` etc.
      * ``sync`` (default) - Contains synchronous cache adapters with simple alias names, e.g. ``redis``, ``memory``, ``memcached``
      * ``asyncio`` / ``async`` - Contains AsyncIO cache adapters with simple alias names, e.g. ``redis``, ``memory``, ``memcached``
    
    Examples::
        
        >>> _MemcachedCache = import_adapter('memcached')
        >>> _AsyncRedisCache = import_adapter('redis', 'asyncio')
        >>> _SqliteCache = import_adapter('SqliteCache')
        
        >>> mc = _MemcachedCache()
        >>> mc.set('hello', 'world')
        >>> mc['hello']
        'world'
    
    :param str name: The dictionary key / name for the adapter, e.g. ``memcached`` / ``sqlite3``, or full adapter class name
                      such as ``AsyncMemoryCache`` if ``cat`` is ``'shared'``
    
    :param str cat: (default: ``'sync'``) One of: ``'sync'``, ``'asyncio'``, ``'async'``, or ``'shared'``
    
    :param bool fallback_shared: (default: ``True``) When set to ``True``, if ``name`` can't be found in the current ``cat`` category,
                                 check if it exists in the ``shared`` category - if it does, then retrieve the module path from that.
    
    :raises KeyError: When the adapter name ``name`` isn't found in the category (nor in ``shared`` if ``fallback_shared`` is enabled)
    :raises AttributeError: When the category ``cat`` isn't found in :attr:`.ADAPTER_MAP`
    
    :return Type[Union[CacheAdapter, AsyncCacheAdapter]] adp_class: The uninstantiated class for the cache adapter that was imported.
    """
    cat = 'asyncio' if cat == 'async' else cat.lower()
    dkeys = list(_AM.keys())
    if cat not in dkeys:
        raise AttributeError(f"Invalid category. You must specify one of the following categories (aka 'cat'): {dkeys}")
    modx: dict = _AM[cat]
    try:
        mpath = modx[name if name in modx else name.lower()]
    except (KeyError, IndexError) as e:
        if fallback_shared and (name.endswith('Cache') or name.endswith('Adapter')) and name in _AM['shared']:
            log.info(f"Adapter name '{name}' wasn't found in category '{cat}' - but WAS found in the shared category, and "
                     f"fallback_shared is enabled. Falling back to: _AM['shared']['{name}']")
            mpath = _AM['shared'][name]
        else:
            raise KeyError(f"Adapter name '{name}' not found in category '{cat}'. Available adapters in '{cat}': {modx.keys()}")
    return import_adapter_path(mpath)
    

class CacheWrapper(object):
    """
    **CacheWrapper** is a small class designed to wrap an instance of :class:`.CacheAdapter` and allow
    the adapter to be switched out at any time, using the static class attribute :py:attr:`.cache_instance`.
    
    This class is used for the singleton global variable :py:attr:`.cached`
    
    For convenience, if :py:attr:`.cache_instance` isn't set-up when something makes an adapter-dependant call, then
    the adapter class in :py:attr:`.default_adapter` will be instantiated and stored in :py:class:`.cache_instance`
    
        >>> # Using the ``: CacheAdapter`` type hinting will allow most IDEs to treat the wrapper as if it were
        >>> # a normal CacheAdapter child class, thus showing appropriate completion / usage warnings
        >>> c: CacheAdapter = CacheWrapper()
        >>> c.set('hello', 'world')
        >>> c['hello']
        'world'
    
    You can replace the cache adapter singleton using the module function :py:func:`.adapter_set` (recommended)
    
        >>> from privex.helpers import cache, CacheWrapper
        >>> cache.adapter_set(cache.MemoryCache())   # Set the current adapter for both the cache module, and wrapper.

    If you only plan to use this wrapper, then you can use :py:meth:`.set_adapter` to update the current cache adapter
    instance.
    
        >>> CacheWrapper.set_adapter(cache.MemoryCache())  # Set the adapter only for the wrapper (aka ``cached``)
    
    """
    cache_instance: Optional[INSCacheAdapter] = None
    """Holds the singleton instance of a :class:`.CacheAdapter` implementation"""
    
    default_adapter: Union[Type[CLSCacheAdapter], str] = settings.DEFAULT_CACHE_ADAPTER
    """The default adapter class to instantiate if :py:attr:`.cache_instance` is ``None``"""

    instance_args = []
    instance_kwargs = {}
    is_adapter_async: bool = False
    
    max_context_layers: int = 1
    _ctx_tracker: Optional[LayeredContext] = None
    
    @classmethod
    def get_context_tracker(cls, reset: bool = False) -> LayeredContext:
        return cls.reset_context_tracker() if reset or not cls._ctx_tracker else cls._ctx_tracker
    
    @classmethod
    def reset_context_tracker(cls) -> LayeredContext:
        cls._ctx_tracker = LayeredContext(cls.cache_instance, max_layers=cls.max_context_layers)
        return cls._ctx_tracker

    @classmethod
    def get_adapter(cls, default: CLSCacheAdapter = default_adapter, *args, **kwargs) -> INSCacheAdapter:
        """
        Attempt to get the singleton cache adapter from :py:attr:`.cache_instance` - if the instance is ``None``, then
        attempt to instantiate ``default()``

        If any ``*args`` or ``**kwargs`` are passed, they will be passed through to ``default(*args, **kwargs)`` so
        that any necessary configuration parameters can be passed to the class.
        """
        if not cls.cache_instance:
            cls.instance_args, cls.instance_kwargs = list(args), dict(kwargs)
            if isinstance(default, str): default = import_adapter(default, 'asyncio' if cls.is_adapter_async else 'sync')
            cls.set_adapter(default(*args, **kwargs))
        return cls.cache_instance

    @classmethod
    def set_adapter(cls, adapter: ANYCacheAdapter, *args, **kwargs) -> INSCacheAdapter:
        cls.instance_args, cls.instance_kwargs = list(args), dict(kwargs)
        cls.cache_instance = None
        if isinstance(adapter, str): adapter = import_adapter(adapter, 'asyncio' if cls.is_adapter_async else 'sync')

        cls.cache_instance = cls.get_adapter(adapter, *args, **kwargs) if isclass(adapter) else adapter
        cls.reset_context_tracker()
        return cls.cache_instance

    @classmethod
    def reset_adapter(cls, default: CLSCacheAdapter = default_adapter, *args, **kwargs) -> INSCacheAdapter:
        """
        Re-create the adapter instance at :attr:`.cache_instance` with the same adapter class (assuming it's set)
        """
        adp = cls.get_adapter(default, *args, **kwargs)
        n_args, n_kwargs = empty_if(args, cls.instance_args, itr=True), {**cls.instance_kwargs, **kwargs}
        c = adp.__class__
        cls.cache_instance = c.__init__(c(), *n_args, **n_kwargs)
        return cls.cache_instance

    def __getattr__(self, item):
        if hasattr(super(), item):
            return getattr(self, item)
        
        @awaitable
        def _wrapper(*args, **kwargs):
            with CacheWrapper.get_adapter() as a:
                return getattr(a, item)(*args, **kwargs)
        
        return _wrapper
    
    @awaitable
    def __getitem__(self, item):
        try:
            with CacheWrapper.get_adapter() as a:
                return a.get(key=item, fail=True)
        except CacheNotFound:
            raise KeyError(f'Key "{item}" not found in cache.')

    @awaitable
    def __setitem__(self, key, value):
        with CacheWrapper.get_adapter() as a:
            return a.set(key=key, value=value)

    async def __aenter__(self):
        """Pass-through to :meth:`.cache_instance.__aenter__` instance AsyncIO context enter method"""
        self.get_adapter()  # Make sure cache_instance exists by calling get_adapter (which will set the default if it doesn't)
        return await self.get_context_tracker().aenter()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Pass-through to :meth:`.cache_instance.__aexit__` instance AsyncIO context exit method"""
        return await self.get_context_tracker().aexit(exc_type, exc_val, exc_tb)

    def __enter__(self):
        """Pass-through to :meth:`.cache_instance.__enter__` instance context enter method"""
        self.get_adapter()  # Make sure cache_instance exists by calling get_adapter (which will set the default if it doesn't)
        return self.get_context_tracker().enter()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Pass-through to :meth:`.cache_instance.__exit__` instance context exit method"""
        return self.get_context_tracker().exit(exc_type, exc_val, exc_tb)


cached: Union[CacheAdapter, CacheWrapper] = CacheWrapper()
"""
This module attribute acts as a singleton, containing an instance of :class:`.CacheWrapper` which is designed
to allow painless usage of this caching module.
"""


class AsyncCacheWrapper(CacheWrapper):
    """
    For applications/packages which are primarily AsyncIO, :class:`.CacheWrapper` can cause problems such as the
    common ``event loop is already running`` - and unfortunately, :mod:`nest_asyncio` isn't always able to fix it.
    
    This wrapper - :class:`.AsyncCacheWrapper` is ONLY compatible with AsyncIO code, and holds a different
    adapter instance in :attr:`.cache_instance` than :class:`.CacheWrapper`, as it expects the adapter instance to
    always be based on :class:`.AsyncCacheAdapter` - along with the fact it generally returns coroutines.
    
        >>> from privex.helpers.cache import async_cached, async_adapter_set, AsyncRedisCache
        >>> async_adapter_set(AsyncRedisCache())
        >>> c = async_cached      # We alias async_cached to 'c' for convenience.
        >>> c['hello'] = 'world'  # We can set cache keys as if the wrapper was a dictionary
        >>> c['hello']            # Similarly we can also get keys like normal
        'world'
        >>> await c['hello']      # It's safest to use ``await`` when getting keys within an async context
        'world'
        >>> await c.get('hello')  # We can also use the coroutines .get, .set, .get_or_set etc.
        'world'
    
    
    """
    cache_instance: AsyncCacheAdapter = None
    """Holds the singleton instance of a :class:`.AsyncCacheAdapter` implementation"""
    
    instance_args = []
    instance_kwargs = {}
    is_adapter_async: bool = True
    
    default_adapter: Union[Type[AsyncCacheAdapter], str] = settings.DEFAULT_ASYNC_CACHE_ADAPTER
    """The default adapter class to instantiate if :py:attr:`.cache_instance` is ``None``"""

    max_context_layers: int = 1
    _ctx_tracker: Optional[LayeredContext] = None
    
    @classmethod
    def get_adapter(cls, default: Union[Type[AsyncCacheAdapter], str] = default_adapter, *args, **kwargs) -> AsyncCacheAdapter:
        
        # if not cls.cache_instance:
        #     cls.instance_args, cls.instance_kwargs = list(args), dict(kwargs)
        #     cls.cache_instance = default(*args, **kwargs)
        return super(AsyncCacheWrapper, cls).get_adapter(default, *args, **kwargs)

    @classmethod
    def set_adapter(cls, adapter: Union[AsyncCacheAdapter, str], *args, **kwargs) -> AsyncCacheAdapter:
        return super(AsyncCacheWrapper, cls).set_adapter(adapter, *args, **kwargs)
        # cls.cache_instance = adapter
        # return cls.cache_instance
    
    @classmethod
    def reset_adapter(cls, default: Type[AsyncCacheAdapter] = default_adapter, *args, **kwargs) -> AsyncCacheAdapter:
        """
        Re-create the adapter instance at :attr:`.cache_instance` with the same adapter class (assuming it's set)
        """
        # adp = cls.get_adapter(default, *args, **kwargs)
        # n_args, n_kwargs = empty_if(args, cls.instance_args, itr=True), {**cls.instance_kwargs, **kwargs}
        # c = adp.__class__
        # cls.cache_instance = c.__init__(c(), *n_args, **n_kwargs)
        return super(AsyncCacheWrapper, cls).reset_adapter(default, *args, **kwargs)
    
    def __getattr__(self, item):
        if hasattr(super(), item):
            return getattr(self, item)
        
        async def _wrapper(*args, **kwargs):
            async with AsyncCacheWrapper.get_adapter() as a:
                return await await_if_needed(getattr(a, item)(*args, **kwargs))
        
        return _wrapper

    @awaitable
    def __getitem__(self, item):
        async def _wrapper():
            try:
                async with AsyncCacheWrapper.get_adapter() as a:
                    return await await_if_needed(a.get(key=item, fail=True))
            except CacheNotFound:
                raise KeyError(f'Key "{item}" not found in cache.')
        return _wrapper()

    def __setitem__(self, key, value):
        async def _wrapper():
            async with AsyncCacheWrapper.get_adapter() as a:
                return await await_if_needed(a.set(key=key, value=value))
        return loop_run(_wrapper())
    
    # async def __aenter__(self):
    #     """Pass-through to :meth:`.cache_instance.__aenter__` instance AsyncIO context enter method"""
    #     return await self.get_adapter().__aenter__()
    #
    # async def __aexit__(self, exc_type, exc_val, exc_tb):
    #     """Pass-through to :meth:`.cache_instance.__aexit__` instance AsyncIO context exit method"""
    #     return await self.get_adapter().__aexit__(exc_type, exc_val, exc_tb)
    def __enter__(self):
        self.get_adapter()  # Make sure cache_instance exists by calling get_adapter (which will set the default if it doesn't)
        # Because this is an AsyncIO-only cache adapter, generally __enter__ isn't used, instead the AsyncIO `__aenter__` and
        # `__aexit__` are used. So, we can try to convert any attempts to use the classic `with x as y` into `async with x as y`
        # by grabbing the event loop and trying to run `__aenter__` in the AsyncIO event loop.
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.get_context_tracker().aenter())

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Because this is an AsyncIO-only cache adapter, generally __exit__ isn't used, instead the AsyncIO `__aenter__` and
        # `__aexit__` are used. So, we can try to convert any attempts to use the classic `with x as y` into `async with x as y`
        # by grabbing the event loop and trying to run `__aexit__` in the AsyncIO event loop.
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.get_context_tracker().aexit(exc_type, exc_val, exc_tb))


async_cached: Union[AsyncCacheAdapter, AsyncCacheWrapper] = AsyncCacheWrapper()


def async_adapter_set(adapter: Union[AsyncCacheAdapter, str]) -> AsyncCacheAdapter:
    """
    Same as :func:`.adapter_set` but sets ``__STORE['async_adapter']`` instead of ``'adapter'``,
    and sets the adapter for async-only :attr:`.async_cached` ( :class:`.AsyncCacheWrapper` ) instead of
    the dual-sync :attr:`.cached` ( :class:`.CacheWrapper` )
    
    Example::
    
        >>> from privex.helpers.cache import AsyncRedisCache, async_adapter_set
        >>> # Passing an instantiated adapter class
        >>> async_adapter_set(AsyncRedisCache())
        >>> # Passing a simple string adapter name to be imported and instantiated
        >>> async_adapter_set('memcached')          # You can either pass a simple "alias" name
        >>> async_adapter_set('AsyncSqliteCache')   # Or you can pass a full adapter class name
    
    """
    if isinstance(adapter, str):
        c_adapter = import_adapter(adapter, 'asyncio')
        adapter = c_adapter()
    __STORE['async_adapter'] = adapter
    async_cached.set_adapter(adapter)
    return __STORE['async_adapter']


def async_adapter_get(default: Union[Type[AsyncCacheAdapter], str] = settings.DEFAULT_ASYNC_CACHE_ADAPTER) -> AsyncCacheAdapter:
    """Same as :func:`.adapter_get` but gets ``__STORE['async_adapter']`` instead of ``'adapter'``"""
    if 'async_adapter' not in __STORE or __STORE['async_adapter'] is None:
        if not default:
            raise NotConfigured('No async cache adapter has been configured for privex.helpers.cache!')
        if isinstance(default, str):
            default = import_adapter(default, 'asyncio')
        async_adapter_set(default())
    return __STORE['async_adapter']


def adapter_set(adapter: Union[CacheAdapter, str]) -> CacheAdapter:
    """
    Set the global cache adapter instance to ``adapter`` - which should be an instantiated adapter class which
    implements :class:`.CacheAdapter`
    
    **Example**::
    
        >>> from privex.helpers import cache
        >>> cache.adapter_set(cache.MemoryCache())   # Passing an instantiated adapter class
        >>> cache.adapter_set('redis')               # Alternatively, you can pass a string alias name of an adapter
        >>> cache.adapter_set('SqliteCache')         # Or the full class name of an adapter
    
    :param CacheAdapter adapter: An instance of a class which implements :class:`.CacheAdapter` for global use.
    :return CacheAdapter adapter: A reference to your adapter from ``__STORE['adapter']``
    """
    if isinstance(adapter, str):
        c_adapter = import_adapter(adapter, 'sync')
        adapter = c_adapter()
    __STORE['adapter'] = adapter
    cached.set_adapter(adapter)
    
    return __STORE['adapter']


def adapter_get(default: Union[Type[CacheAdapter], str] = settings.DEFAULT_CACHE_ADAPTER) -> CacheAdapter:
    """
    Get the global cache adapter instance. If there isn't one, then by default this function will initialise
    :class:`.MemoryAdapter` and set it as the global cache adapter.
    
    To set the global cache adapter instance, use :py:func:`.adapter_set`
    
    To use a different fallback class, pass a class name which implements :class:`.CacheAdapter` like so:
    
        >>> adapter_get(default=MemoryCache)
    
    :param default:
    :return:
    """
    if 'adapter' not in __STORE or __STORE['adapter'] is None:
        if not default:
            raise NotConfigured('No cache adapter has been configured for privex.helpers.cache!')
        if isinstance(default, str):
            default = import_adapter(default, 'sync')
        __STORE['adapter'] = default()
    return __STORE['adapter']


def get(key: str, default: Any = None, fail: bool = False) -> Any:
    """
    Return the value of cache key ``key``. If the key wasn't found, or it was expired, then ``default`` will be
    returned.
    
    Optionally, you may choose to pass ``fail=True``, which will cause this method to raise :class:`.CacheNotFound`
    instead of returning ``default`` when a key is non-existent / expired.
    
    :param str key: The cache key (as a string) to get the value for, e.g. ``example:test``
    :param Any default: If the cache key ``key`` isn't found / is expired, return this value (Default: ``None``)
    :param bool fail: If set to ``True``, will raise :class:`.CacheNotFound` instead of returning ``default``
                      when a key is non-existent / expired.
    
    :raises CacheNotFound: Raised when ``fail=True`` and ``key`` was not found in cache / expired.
    
    :return Any value: The value of the cache key ``key``, or ``default`` if it wasn't found.
    """
    a = adapter_get()
    return a.get(key=key, default=default, fail=fail)


def set(key: str, value: Any, timeout: Optional[int] = DEFAULT_CACHE_TIMEOUT):
    """
    Set the cache key ``key`` to the value ``value``, and automatically expire the key after ``timeout`` seconds
    from now.
    
    If ``timeout`` is ``None``, then the key will never expire (unless the cache implementation loses it's
    persistence, e.g. memory caches with no disk writes).
    
    :param str key: The cache key (as a string) to set the value for, e.g. ``example:test``
    :param Any value: The value to store in the cache key ``key``
    :param int timeout: The amount of seconds to keep the data in cache. Pass ``None`` to disable expiration.
    """
    a = adapter_get()
    return a.set(key=key, value=value, timeout=timeout)


def get_or_set(key: str, value: Union[Any, callable], timeout: int = DEFAULT_CACHE_TIMEOUT) -> Any:
    """
    Attempt to return the value of ``key`` in the cache. If ``key`` doesn't exist or is expired, then it will be
    set to ``value``, and ``value`` will be returned.
    
    The ``value`` parameter can be any standard type such as ``str`` or ``dict`` - or it can be a callable
    function / method which returns the value to set and return.
    
    **Basic Usage**::
    
        >>> from privex.helpers import cache as c
        >>> c.get('testing')
        None
        >>> c.get_or_set('testing', 'hello world')
        'hello world'
        >>> c.get('testing')
        'hello world'
    
    **Set and get the value from a function if ``key`` didn't exist / was expired**::
    
        >>> def my_func(): return "hello world"
        >>> c.get_or_set('example', my_func)
        'hello world'
        >>> c.get('example')
        'hello world'
    
    :param str key: The cache key (as a string) to get/set the value for, e.g. ``example:test``
    :param Any value: The value to store in the cache key ``key``. Can be a standard type, or a callable function.
    :param int timeout: The amount of seconds to keep the data in cache. Pass ``None`` to disable expiration.
    :return Any value: The value of the cache key ``key``, or ``value`` if it wasn't found.
    """
    a = adapter_get()
    return a.get_or_set(key=key, value=value, timeout=timeout)


def remove(*key: str) -> bool:
    """
    Remove one or more keys from the cache.

    If all cache keys existed before removal, ``True`` will be returned. If some didn't exist (and thus couldn't
    remove), then ``False`` will be returned.

    :param str key: The cache key(s) to remove
    :return bool removed: ``True`` if ``key`` existed and was removed
    :return bool removed: ``False`` if ``key`` didn't exist, and no action was taken.
    """
    a = adapter_get()
    return a.remove(*key)


def update_timeout(key: str, timeout: int = DEFAULT_CACHE_TIMEOUT) -> Any:
    """
    Update the timeout for a given ``key`` to ``datetime.utcnow() + timedelta(seconds=timeout)``

    This method allows keys which are already expired, allowing expired cache keys to have their timeout
    extended **after** expiry.

    **Example**::

        >>> from privex.helpers import cache
        >>> from time import sleep
        >>> cache.set('example', 'test', timeout=60)
        >>> sleep(70)
        >>> cache.update_timeout('example', timeout=60)   # Reset the timeout for ``'example'`` to ``now + 60 seconds``
        >>> cache.get('example')
        'test'

    :param str key: The cache key to update the timeout for
    :param int timeout: Reset the timeout to this many seconds from ``datetime.utcnow()``
    :raises CacheNotFound: Raised when ``key`` was not found in cache (thus cannot extend timeout)
    :return Any value: The value of the cache key
    """
    a = adapter_get()
    return a.update_timeout(key=key, timeout=timeout)




"""
Helper functions and classes to ease the use :class:`.Thread`'s with python's :mod:`threading` library

Utilities for working with :class:`.Event`
------------------------------------------


Classes
^^^^^^^

    ==============================   ==================================================================================================
    Class                            Description
    ==============================   ==================================================================================================
    :class:`.BetterEvent`            **BetterEvent** is a sub-class of :class:`.Event` with more flexibility + features
    :class:`.InvertibleEvent`        (this is just an alias for :class:`.BetterEvent`)
    ==============================   ==================================================================================================


Functions
^^^^^^^^^

    ==============================   ==================================================================================================
    Function                         Description
    ==============================   ==================================================================================================
    :func:`.event_multi_wait`        Allows waiting for more than one :class:`.Event` / :class:`.BetterEvent` at once
    :func:`.event_multi_wait_all`    A wrapper function for :func:`.event_multi_wait` with ``trigger='and'`` as default
    :func:`.event_multi_wait_any`    A wrapper function for :func:`.event_multi_wait` with ``trigger='or'`` as default
    ==============================   ==================================================================================================


Utilities for working with :class:`.Lock` thread locks
------------------------------------------------------

In the Python standard library, you can only use context management directly against a :class:`.Lock` object, which means
you're unable to specify things such as a timeout, whether or not to block, nor a built-in option to request an exception
to be raised if the lock can't be acquired.

And thus, :func:`.lock_acquire_timeout` was created - to solve all of the above problems, in one easy to use context management
function :)

The :func:`.lock_acquire_timeout` function - a context manager ( ``with lock_acquire_timeout(lock)`` ), is designed to allow use of
context management with standard :class:`threading.Lock` objects, with the ability to specify important parameters such as:
 
 * whether or not to **block** while acquiring the lock
 * an optional timeout - so that it gives up waiting for the ``lock.acquire`` after so many seconds
 * whether to raise :class:`.LockWaitTimeout` if the ``acquire`` times out instead of returning ``None``

Functions
^^^^^^^^^

    ==============================   ==================================================================================================
    Function                         Description
    ==============================   ==================================================================================================
    :func:`.lock_acquire_timeout`    Flexible context manager for acquiring :class:`.Locks`'s ``with lock_acquire_timeout(lock)``
    ==============================   ==================================================================================================

Utilities for working with :class:`.Thread` thread objects
----------------------------------------------------------

Using :class:`.SafeLoopThread` for looping threads with :class:`queue.Queue`'s + stop/pause support
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

First example - we'll create a sub-class of :class:`.SafeLoopThread` called ``MyThread``::

    >>> # Create a sub-class of SafeLoopThread, and implement a loop() method
    >>> class MyThread(SafeLoopThread):
    ...     loop_sleep = 2    # 'run' will wait this many seconds between each run of your loop(). set to 0 to disable loop sleeps
    ...     def loop(self):
    ...         print("I'm looping!")
    ...
    >>> t = MyThread()   # Construct the class

Once we start the thread, we'll see that ``I'm looping!`` will be printed about once every 2 seconds, since that's what
we set ``loop_sleep`` to::

    >>> t.start()
    I'm looping!
    I'm looping!
    I'm looping!

Using :meth:`.SafeLoopThread.emit_pause` - we can pause the loop, which will silence the ``I'm looping!`` messages::

    >>> t.emit_pause()
    >>> # No output because the loop is now paused

To start the loop again, we can simply unpause it with :meth:`.SafeLoopThread.emit_unpause`::

    >>> t.emit_unpause()
    I'm looping!
    I'm looping!
    I'm looping!

Once we're done with ``MyThread``, unlike a normal :class:`.Thread`, we can ask the thread to shutdown gracefully
using :meth:`.SafeLoopThread.emit_stop` like so::

    >>> t.is_alive()    # First we'll confirm the thread is still running
    True
    >>> t.emit_stop()
    I'm looping!
    >>> # The .loop method will finish it's current iteration (unless you add additional ``should_stop`` checks in the loop)
    >>> # and then shutdown the thread by returning from ``.run``
    >>> t.is_alive()    # We can now see that the thread has shutdown as we requested it to.
    False



Classes
^^^^^^^

    ==============================   ==================================================================================================
    Class                            Description
    ==============================   ==================================================================================================
    :class:`.StopperThread`          A :class:`.Thread` base class which allows you easily add stop/pause support to your own threads
    :class:`.SafeLoopThread`         A :class:`.StopperThread` based class which runs ``.loop`` in a loop, with stop/start support
    ==============================   ==================================================================================================



"""
import queue
import threading
import time
from threading import Lock, Event
from contextlib import contextmanager
from datetime import datetime
from typing import Union, Optional, List

from privex.helpers.exceptions import LockWaitTimeout, EventWaitTimeout

import logging

log = logging.getLogger(__name__)

__all__ = [
    'lock_acquire_timeout', 'BetterEvent', 'InvertibleEvent',
    'StopperThread', 'SafeLoopThread',
    'event_multi_wait', 'event_multi_wait_all', 'event_multi_wait_any'
]


@contextmanager
def lock_acquire_timeout(lock: Lock, timeout: Union[int, float] = 10, fail=False, block=True):
    """
    A context manager (``with lock_acquire_timeout(mylock) as locked:``) for acquiring thread locks and waiting for them to be released,
    and giving up if the lock isn't released within ``timeout``.
    
    Yields a boolean in the ``with`` context which is ``True`` if the :class:`threading.Lock` was acquired within ``timeout``,
    or ``False`` if it wasn't.
    
        >>> from privex.helpers import lock_acquire_timeout
        >>> from threading import Lock
        >>>
        >>> my_lock = Lock()
        >>>
        >>> def some_func():
        ...     print("attempting to acquire a lock on 'my_lock'... will wait up to 30 secs...")
        ...     with lock_acquire_timeout(my_lock, timeout=30) as locked:
        ...         if not locked:
        ...             raise Exception("Failed to acquire 'my_lock' after waiting 30 seconds!")
        ...         print("successfully acquired a lock on 'my_lock'")
        ...     print("finished. my_lock should have been automatically released.")
    
    Original written by "robbles" on StackOverflow: https://stackoverflow.com/a/16782391/2648583
    
    :param Lock lock:           The :class:`threading.Lock` object to attempt to acquire a lock on
    :param int|float timeout:   The amount of seconds to wait for ``lock`` to be released if it's already locked
    :param bool fail:           (Default: ``False``) If this is ``True``, will raise :class:`.LockWaitTimeout` if we fail to
                                acquire the lock ``lock`` within ``timeout`` seconds.
    :param bool block:          If this is set to ``False``, ``timeout`` will be nulled and a non-blocking acquire will be done.
    :raises LockWaitTimeout:    When ``fail`` is ``True`` and we fail to acquire ``lock`` within ``timeout`` seconds.
    """
    timeout = timeout if block else None
    fail_msg = f"after waiting {timeout} seconds." if block else "(non-blocking acquire)"
    if block:
        log.debug("acquiring lock with timeout %s", timeout)
        result = lock.acquire(blocking=block, timeout=timeout)
    else:
        log.debug("non-blocking mode requested. ignoring timeout - using non-blocking acquire")
        result = lock.acquire(blocking=block)
    if not result and fail:
        raise LockWaitTimeout(f"Failed to acquire lock '{lock}' {fail_msg}")
    log.debug("post-acquire - result was %s after timeout %s - handing to context...", result, timeout)

    yield result
    log.debug("lock context finished")
    if result:
        lock.release()


class BetterEvent(Event):
    """
    :class:`.BetterEvent` (alias :attr:`.InvertibleEvent`) is a more flexible version of :class:`threading.Event`, which
    adds many new capabilities / features on top of :class:`threading.Event`:
    
     * The ``wait_on`` constructor parameter allows you to choose what flag states that the standard :meth:`.wait` will
       trigger upon:
       
        * ``'set'`` - the default - works like :class:`threading.Event`, :meth:`.wait` only triggers when the event is in
          the "set" state, i.e. :attr:`._flag` is ``True``
        * ``'clear'`` - opposite of the default - works opposite to :class:`threading.Event`, :meth:`.wait` only triggers when the event
          is in the "clear" state, i.e. :attr:`._flag` is ``False``
        * ``'both'`` - In the ``both`` setting, :meth:`.wait` will simply wait until :attr:`._flag` is changed, whether from
          ``set`` to ``clear``, or ``clear`` to ``set``.
          This wait_on setting only works as long as ``notify_set`` and ``notify_clear`` are set to ``True``
     
     * The ``default`` constructor parameter allows you to choose whether the event starts as "cleared" (``False`` - default), or
       "set" (``True``), which is useful when using some of the alternative ``wait_on`` settings.
     
     * New ``fail`` parameter for :meth:`.wait`, :meth:`.wait_set` and :meth:`.wait_clear` - when this is set to ``True``,
       the method will raise :class:`.EventWaitTimeout` when the timeout is hit, instead of just returning ``False``.
    
     * New :meth:`.wait_set` method, this works like the classic :class:`threading.Event` ``wait`` method - it's only triggered
       when :attr:`._flag` is set to ``True`` (set) - no matter what ``wait_on`` setting is active.
     
     * New :meth:`.wait_clear` method, this works opposite to the classic :class:`threading.Event` ``wait`` method - it's only triggered
       when :attr:`._flag` is set to ``False`` (cleared) - no matter what ``wait_on`` setting is active.
    
    
    **Example Usage**
    
    Below is a very simple thread class using :class:`.SafeLoopThread` which uses a :class:`.BetterEvent` so we can signal
    when it can start running, and when it's allowed to restart itself::
    
        >>> from privex.helpers import BetterEvent, SafeLoopThread
        >>>
        >>> class MyThread(SafeLoopThread):
        ...     def __init__(self, *args, trig, **kwargs):
        ...         self.trig = trig
        ...         super().__init__(*args, **kwargs)
        ...     def loop(self):
        ...         print("Waiting for trig to become set before doing stuff...")
        ...         self.trig.wait()   # Same behaviour as threading.Event.wait - waits for trig.set()
        ...         print("trig is set. doing stuff...")
        ...         print("finished doing stuff.")
        ...         print("Waiting for trig to become clear before restarting loop...")
        ...         self.trig.wait_clear()  # Unlike threading.Event, BetterEvent allows waiting for the "clear" signal
        ...
        >>> evt = BetterEvent(name='My Event')
        >>> t = MyThread(trig=evt)
        >>> t.start()
        Waiting for trig to become set before doing stuff...
        >>> evt.set()   # We flip evt (trig) to "set", which notifies MyThread it can proceed.
        trig is set. doing stuff...
        finished doing stuff.
        Waiting for trig to become clear before restarting loop...
        >>> evt.clear()  # Unlike threading.Event, we can "clear" the event, and MyThread will detect the "clear" signal instantly.
        Waiting for trig to become set before doing stuff...
        >>> evt.set()    # The loop restarted. Now we can flip trig back to "set"
        trig is set. doing stuff...
        finished doing stuff.
        Waiting for trig to become clear before restarting loop...
    
    
    """
    wait_on: str
    name: Optional[str]
    notify_set: bool
    notify_clear: bool
    
    _flag: bool
    _cond: threading.Condition
    
    def __init__(self, wait_on: str = 'set', name: str = None, default: bool = False, notify_set=True, notify_clear=True):
        """
        
        :param wait_on: ``set`` (default) - :meth:`.wait` triggers when event is "set". ``clear`` - :meth:`.wait` triggers when
                        event is "clear". ``both`` - :meth:`.wait` triggers when the event state flips from "set" to "clear" or vice versa.
        
        :param str name: An optional name to identify this event
        
        :param default: The default state of the event. Either ``False`` for "clear", or ``True`` for "set"
        
        :param notify_set: Whether to notify state listeners :meth:`.wait` :meth:`.wait_set` :meth:`.wait_clear`
                           when :meth:`.set` is called and the state changes from "clear" to "set"
                           
        :param notify_clear: Whether to notify state listeners :meth:`.wait` :meth:`.wait_set` :meth:`.wait_clear`
                             when :meth:`.clear` is called and the state changes from "set" to "clear"
        """
        super().__init__()
        self._cond = threading.Condition(Lock())
        self._flag = default
        self.name = name
        self.notify_set = notify_set
        self.notify_clear = notify_clear
        self.wait_on = wait_on.lower()
        if self.wait_on not in ['set', 'clear', 'both']:
            raise AttributeError("wait_on must be either 'set', 'clear' or 'both'")
        
    def set(self):
        with self._cond:
            if self._flag: return False
            self._flag = True
            if self.notify_set: self._cond.notify_all()
            return True

    def clear(self):
        with self._cond:
            if not self._flag: return False
            self._flag = False
            if self.notify_clear: self._cond.notify_all()
            return True
    
    def _fail_or_signal(self, signal, timeout, fail=False):
        if fail and not signal:
            raise EventWaitTimeout(f"Timed out after waiting {timeout} seconds for event signal (event: {self})")
        return signal
    
    def wait_set(self, timeout: Optional[float] = None, fail=False) -> bool:
        """Wait until :attr:`._flag` is ``True``"""
        timeout = None if timeout is None else float(timeout)
        with self._cond:
            signaled = self._flag
            if not signaled:
                signaled = self._cond.wait(timeout)
            return self._fail_or_signal(signaled, timeout, fail=fail)

    def wait_clear(self, timeout: Optional[float] = None, fail=False) -> bool:
        """Wait until :attr:`._flag` is ``False``"""
        timeout = None if timeout is None else float(timeout)
        with self._cond:
            flag, signaled = self._flag, True
            if flag:
                signaled = self._cond.wait(timeout)
            return self._fail_or_signal(signaled, timeout, fail=fail)

    def wait(self, timeout: Optional[Union[int, float]] = None, fail=False) -> bool:
        """
        Multi-purpose ``wait`` method which works similarly to :meth:`threading.Event.wait` but with some extra features.
        
        This method's behaviour will vary depending on what the :attr:`.wait_on` setting is set to:
        
        * ``'set'`` - the default - works like :class:`threading.Event`, :meth:`.wait` only triggers when the event is in
          the "set" state, i.e. :attr:`._flag` is ``True``
        
        * ``'clear'`` - opposite of the default - works opposite to :class:`threading.Event`, :meth:`.wait` only triggers when the event
          is in the "clear" state, i.e. :attr:`._flag` is ``False``
        
        * ``'both'`` - In the ``both`` setting, :meth:`.wait` will simply wait until :attr:`._flag` is changed, whether from
          ``set`` to ``clear``, or ``clear`` to ``set``.
          This wait_on setting only works as long as ``notify_set`` and ``notify_clear`` are set to ``True``
        
        :param bool fail: If ``True``, raise :class:`.EventWaitTimeout` if ``timeout`` was reached while waiting for the event to change.
        :param float timeout: Maximum amount of time to wait before giving up (``None`` to disable timeout)
        :return bool signal: This method returns the internal flag on exit, so it will always return
                             ``True`` except if a timeout is given and the operation times out.
        """
        timeout = None if timeout is None else float(timeout)
        with self._cond:
            flag, signaled = self._flag, True
            if self.wait_on == 'set' and not flag:
                signaled = self._cond.wait(timeout)
            if self.wait_on == 'clear' and flag:
                signaled = self._cond.wait(timeout)
            if self.wait_on == 'both':
                signaled = self._cond.wait(timeout)
            return self._fail_or_signal(signaled, timeout, fail=fail)
    
    def __str__(self):
        return f"<BetterEvent name='{self.name}' wait_on='{self.name}' status='{'set' if self._flag else 'clear'}' >"
    
    def __repr__(self):
        return self.__str__()


InvertibleEvent = BetterEvent


class StopperThread(threading.Thread):
    """
    A :class:`threading.Thread` thread sub-class which implements :class:`.BetterEvent` events allowing you to signal
    the thread when you want it to shutdown or pause.
    
    You must check :attr:`.should_stop` / :attr:`.should_run` and :attr:`.should_pause` within your thread run body
    to detect when your thread needs to shutdown / pause.
    """
    stop_events: List[threading.Event]
    
    def __init__(self, *args, default_stop=False, default_pause=False, stop_events=None, pause_events=None, **kwargs):
        stop_events = [] if stop_events is None else stop_events
        pause_events = [] if pause_events is None else pause_events
        self.ev_stop = BetterEvent(default=default_stop, name=f'{self.__class__.__name__} ev_stop')
        self.ev_pause = BetterEvent(default=default_pause, name=f'{self.__class__.__name__} ev_pause')
        self.stop_events = [self.ev_stop] + stop_events
        self.pause_events = [self.ev_pause] + pause_events
        super().__init__(*args, **kwargs)

    @property
    def should_pause(self) -> bool:
        return any([ev.is_set() for ev in self.pause_events])
    
    @property
    def should_stop(self) -> bool:
        return any([ev.is_set() for ev in self.stop_events])
    
    @property
    def should_run(self) -> bool:
        return not self.should_stop
    
    def emit_stop(self):
        return self.ev_stop.set()
    
    def emit_start(self):
        return self.ev_stop.clear()

    def emit_pause(self):
        return self.ev_pause.set()

    def emit_unpause(self):
        return self.ev_pause.clear()


class SafeLoopThread(StopperThread):
    loop_sleep: float = 0.1
    pause_sleep: float = 0.5
    
    def __init__(self, *args, default_stop=False, default_pause=False, **kwargs):
        """
        This is a simple loop thread class which uses :class:`.StopperThread` to allow an application to signal to the thread
        that it should shutdown or temporarily pause the loop.
        
        Note that if you have a long loop which takes a long time to run, you should regularly check :attr:`.should_stop`
        and :attr:`.should_pause` within your loop, so that the loop can shutdown or pause quickly when you request it to.
        
        Example usage::
        
            >>> # Create a sub-class of SafeLoopThread, and implement a loop() method
            >>> class MyThread(SafeLoopThread):
            ...     loop_sleep = 2    # 'run' will wait this many seconds between each run of your loop(). set to 0 to disable loop sleeps
            ...     def loop(self):
            ...         print("I'm looping!")
            ...
            >>> t = MyThread(default_pause=True)
            >>> t.start()
            >>> # No output because default_pause was set to True. To start the loop, we'll unpause it
            >>> t.emit_unpause()
            I'm looping!
            I'm looping!
            I'm looping!
            >>> t.emit_pause()
            >>> # We can re-pause the loop from our main thread, and the output stops again.
            >>> # If we unpause the thread again, we can also ask the thread to stop looping and shutdown using emit_stop
            >>> t.emit_unpause()
            I'm looping!
            I'm looping!
            >>> t.emit_stop()
            >>> t.is_alive()
            False
            >>> # Because we sent a STOP event, 'run' stopped looping, and returned.
            >>> # To be able to start the same thread object again, we'd need to clear the STOP event first
            >>> t.emit_start()  # Clear the STOP event we set with emit_stop
            >>> t.start()       # Start the thread again
            I'm looping!
            I'm looping!
        
        
        :param args:
        :param default_stop:  If ``True``, will trigger the event :attr:`.ev_stop` in the constructor, which will prevent the thread
                              from being able to start until :attr:`.ev_stop` is cleared e.g. via :meth:`.emit_start`
        :param default_pause:  If ``True``, will trigger the event :attr:`.ev_pause` in the constructor, which will cause 'run'
                              to wait until :attr:`.ev_pause` is cleared e.g. via :meth:`.emit_unpause` before starting the loop
        :param kwargs:
        :key List[Union[Event,BetterEvent]] stop_events: Additional :class:`.Event` or :class:`.BetterEvent` 's which when set, will
                                                         cause :attr:`.should_stop` to become True.
        :key List[Union[Event,BetterEvent]] pause_events: Additional :class:`.Event` or :class:`.BetterEvent` 's which when set, will
                                                          cause :attr:`.should_pause` to become True.
        """
        self.in_queue = queue.Queue()
        self.out_queue = queue.Queue()
        super().__init__(*args, default_stop=default_stop, default_pause=default_pause, **kwargs)
    
    def loop(self):
        raise NotImplemented("child must implement .loop")
    
    def run(self) -> None:
        log.debug(" [%s] Starting loop thread: %s", self.name, self.__class__.__name__)
        while self.should_run:
            if self.should_pause:
                log.debug(" [%s] should_pause is True. Pausing loop until should_pause is cleared.", self.name, self.__class__.__name__)
                while self.should_pause:
                    time.sleep(self.pause_sleep if self.pause_sleep > 0 else 0.5)

            log.debug(" [%s] Running: %s.loop()", self.name, self.__class__.__name__)
            self.loop()
            if self.loop_sleep > 0:
                time.sleep(self.loop_sleep)
        
        return log.debug(" [%s] Finished loop thread %s", self.name, self.__class__.__name__)


_evt_btevt = Union[BetterEvent, Event]

_bl_list_btevt = Optional[Union[bool, List[_evt_btevt]]]


def event_multi_wait(*events: _evt_btevt, trigger='and', event_sleep=0.5, wait_timeout=None, fail=True, **kwargs) -> _bl_list_btevt:
    """
    Wait for multiple :class:`threading.Event` or :class:`.BetterEvent` 's to become "set", or "clear".
    
    For standard :class:`threading.Event` 's - only "set" can be waited on. You must use :class:`.BetterEvent` (generally
    works as a drop-in replacement) to be able to use the ``invert`` or ``invert_indexes`` options.
    
    Basic example::
    
        >>> do_something_else = Event()
        >>> stop_running = Event()
        >>>
        >>> def some_thread():
        ...     # do some stuff...
        ...     # now we wait for further instructions, until either do_something_else or stop_running is signalled:
        ...     event_multi_wait(do_something_else, stop_running, trigger='any')
        ...     if stop_running.is_set(): return False
        ...     if do_something_else.is_set():
        ...         # do something else.


    :param Event|BetterEvent events:    Multiple :class:`threading.Event` references to be waited on.

    :param str trigger:                 To return when ALL events are set, specify one of ``and|all|every``, while to return when ANY
                                        of the specified events are set, specify one of ``or|any|either``

    :param float|int event_sleep:       The maximum amount of time per event check iteration. This is divided by the amount of events
                                        which were passed, so we can use the highly efficient ``event.wait()`` method.

    :param float|int wait_timeout:      The maximum amount of time (in seconds) to wait for all/any of the ``events`` to signal.
                                        Set to ``None`` to disable wait timeout. When timing out, raises :class:`.EventWaitTimeout`
                                        if ``fail=True``, otherwise simply returns ``None``.

    :param bool fail:       When wait_timeout is hit, will raise :class:`.EventWaitTimeout` if ``fail=True``, otherwise
                            will simply return ``None``.
    
    :param kwargs:          Additional settings
    
    :key bool invert:       (Default: ``False``) Wait for ``events`` to become "clear" (``False``). NOTE: This only
                            works with :class:`.BetterEvent` events.
    
    :key list invert_indexes: Wait for the events at these indexes to become "clear" instead of "set".
                              If ``invert`` is set to ``True``, then we'll wait for these indexes to become "set" instead of "clear".
                              NOTE: This only works with :class:`.BetterEvent` events.
    :return bool success:   ``True`` if ``events`` met the ``trigger``, otherwise ``None``
    :return List[_evt_btevt] events: If :class:`.BetterEvents` are passed, and ``trigger`` is "any", then a list of the events
                                     which were set (or if ``invert`` is ``True``, then events that weren't set.
                              
    """
    # Returns True if an event is set the correct direction depending on invert mode and invert_indexes
    def _filter_event(evt: Union[Event, BetterEvent], index: int) -> bool:
        if invert:   # invert mode - "set" = False, "clear" = True (opposite if index in invert_indexes)
            return evt.is_set() if index in invert_indexes else not evt.is_set()
        elif index in invert_indexes:  # normal mode (in invert_indexes) - "set" = False, "clear" = True
            return not evt.is_set()
        else:  # normal mode - "set" = True, "clear" = False
            return evt.is_set()
    
    # Handles deciding whether to use .wait or .wait_clear depending on invert mode and invert_indexes
    def _wait_event(evt: Union[Event, BetterEvent], index: int):
        if invert:
            if index in invert_indexes:  # In invert mode, we wait for events in invert_indexes to be "set"
                evt.wait(sleep_ev)
            else:  # In invert mode, we wait for events to be "clear" when not in invert_indexes
                evt.wait_clear(sleep_ev)
        elif index in invert_indexes:  # In normal mode, we wait for events in invert_indexes to be "clear"
            evt.wait_clear(sleep_ev)
        else:  # If we're not in invert mode, and the event isn't in invert_indexes, then we just use normal .wait()
            evt.wait(sleep_ev)
    
    if len(events) < 1:
        raise AttributeError("event_multi_wait expects one or more events passed as positional arguments...")
    
    invert = kwargs.get('invert', kwargs.get('inverted', False))
    invert_indexes: List[_evt_btevt] = kwargs.get('invert_indexes', kwargs.get('invert_idx', []))
    
    is_better_events = all([isinstance(v, BetterEvent) for v in events])
    if invert and not is_better_events:
        raise AttributeError("You specified invert=True or invert_indexes but one or more passed events are not BetterEvent classes.")
    
    all_is_set = lambda: [_filter_event(evt=v, index=i) for i, v in enumerate(events)]
    event_sleep = float(event_sleep)
    wait_timeout = None if wait_timeout is None else float(wait_timeout)
    
    start_time = datetime.utcnow()
    sleep_ev = event_sleep / len(events)
    if trigger.lower() in ['and', 'all', 'every']:
        while not all(all_is_set()):
            if wait_timeout is not None and (datetime.utcnow() - start_time).total_seconds() > wait_timeout:
                if fail:
                    raise EventWaitTimeout(f"Waited {wait_timeout} seconds for all {len(events)} Event's to signal. Giving up.")
                return None
            for idx, ev in enumerate(events):
                _wait_event(evt=ev, index=idx)
        return True
    
    if trigger.lower() in ['or', 'any', 'either']:
        while not any(all_is_set()):
            if wait_timeout is not None and (datetime.utcnow() - start_time).total_seconds() > wait_timeout:
                if fail:
                    raise EventWaitTimeout(f"Waited {wait_timeout} seconds for any of the {len(events)} Event's to signal. Giving up.")
                return None
            for idx, ev in enumerate(events):
                _wait_event(evt=ev, index=idx)
        
        if is_better_events:
            return [v for i, v in enumerate(events) if _filter_event(evt=v, index=i)]
        return True
    
    raise AttributeError("event_multi_wait expects 'trigger' to be one of: ['and', 'all', 'every', 'or', 'any', 'either']")


def event_multi_wait_all(*events: _evt_btevt, event_sleep=0.5, wait_timeout=None, fail=True, **kwargs) -> bool:
    """
    Wrapper function for :meth:`.event_wait_many` with ``trigger`` defaulting to ``and`` (return when ALL events are triggered)
    """
    return event_multi_wait(*events, trigger='and', event_sleep=event_sleep, wait_timeout=wait_timeout, fail=fail, **kwargs)


def event_multi_wait_any(*events: _evt_btevt, event_sleep=0.5, wait_timeout=None, fail=True, **kwargs) -> bool:
    """
    Wrapper function for :meth:`.event_wait_many` with ``trigger`` defaulting to ``or`` (return when ANY event triggers)
    """
    return event_multi_wait(*events, trigger='or', event_sleep=event_sleep, wait_timeout=wait_timeout, fail=fail, **kwargs)



"""
Various functions/classes which convert/parse objects from one type into another.

**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
import warnings
from datetime import datetime, date
from decimal import Decimal
from typing import Any, Optional, Union, AnyStr

from privex.helpers.exceptions import ValidatorNotMatched
from privex.helpers.types import T
from privex.helpers.collections import Mocker

try:
    from privex.helpers.extras.attrs import AttribDictable
except ImportError as e:
    warnings.warn(f"Failed to import privex.helpers.extras.attrs.AttribDictable - falling back to placeholder type")
    AttribDictable = Mocker.make_mock_class('AttribDictable', instance=False)

try:
    from privex.helpers.collections import Dictable
except ImportError as e:
    warnings.warn(f"Failed to import privex.helpers.collections.Dictable - falling back to placeholder type")
    Dictable = Mocker.make_mock_class('Dictable', instance=False)

try:
    from privex.helpers.collections import DictDataClass
except ImportError as e:
    warnings.warn(f"Failed to import privex.helpers.collections.DictDataClass - falling back to placeholder type")
    DictDataClass = Mocker.make_mock_class('DictDataClass', instance=False)


try:
    import dataclasses
except ImportError as e:
    warnings.warn(f"Failed to import dataclasses - falling back to placeholder type")
    from privex.helpers.mockers import dataclasses

try:
    import attr
    from attr.exceptions import NotAnAttrsClassError
except ImportError as e:
    warnings.warn(f"Failed to import attr - falling back to placeholder type")
    from privex.helpers.mockers import attr


    class NotAnAttrsClassError(Exception):
        pass
    #
    # attr = Mocker(
    #     attributes=dict(
    #         s=mock_decorator,
    #         asdict=lambda obj, dict_factory=dict: dict_factory(obj),
    #         astuple=lambda obj, tuple_factory=tuple: tuple_factory(obj),
    #         validate=lambda obj: False
    #     )
    # )

from privex.helpers.common import empty, is_true, stringify
import logging

log = logging.getLogger(__name__)

MINUTE = 60
HOUR = MINUTE * 60
DAY = HOUR * 24
MONTH = DAY * 30
YEAR = DAY * 365
DECADE = YEAR * 10

SUPPORTED_DT_TYPES = Union[str, bytes, int, datetime, date, AnyStr]


def convert_datetime(d, if_empty=None, fail_empty=False, **kwargs) -> Optional[datetime]:
    """
    Convert the object ``d`` into a :class:`datetime.datetime` object.
    
    If ``d`` is a string or bytes, then it will be parsed using :func:`dateutil.parser.parse`
    
    If ``d`` is an int/float/Decimal, then it will be assumed to be a unix epoch timestamp.
    
    **Examples**::
        
        >>> convert_datetime("2019-01-01T00:00:00Z")          # ISO date/time
        datetime.datetime(2019, 1, 1, 0, 0, tzinfo=tzutc())
        
        >>> convert_datetime("01/JAN/2019 00:00:00.0000")     # Human date/time with month name
        datetime.datetime(2019, 1, 1, 0, 0, tzinfo=tzutc())
        
        >>> convert_datetime(1546300800)                      # Unix timestamp as integer
        datetime.datetime(2019, 1, 1, 0, 0, tzinfo=tzutc())
        
        >>> convert_datetime(1546300800000)                   # Unix timestamp (milliseconds) as integer
        datetime.datetime(2019, 1, 1, 0, 0, tzinfo=tzutc())
    
    :param d: Object to convert into a datetime
    :param if_empty: If ``d`` is empty / None, return this value
    :param bool fail_empty: (Def: ``False``) If this is True, then if ``d`` is empty, raises :class:`AttributeError`
    
    :key datetime.tzinfo tzinfo: (Default: :class:`dateutil.tz.tzutc`) If no timezone was detected by the parser,
                                 use this timezone. Set this to ``None`` to disable forcing timezone-aware dates.
    
    :raises AttributeError: When ``d`` is empty and ``fail_empty`` is set to True.
    :raises dateutil.parser.ParserError: When ``d`` could not be parsed into a date.
    :return datetime converted: The converted :class:`datetime.datetime` object.
    """
    from dateutil.tz import tzutc
    _tzinfo = kwargs.pop('tzinfo', tzutc())
    if isinstance(d, datetime):
        if d.tzinfo is None and _tzinfo is not None:
            d = d.replace(tzinfo=_tzinfo)
        return d
    
    # For datetime.date objects, we first convert them into a string, then we can parse them into a datetime + set tzinfo
    if isinstance(d, date):
        d = str(d)

    d = stringify(d) if isinstance(d, bytes) else d
    
    if isinstance(d, (int, float)):
        return convert_unixtime_datetime(d)
    
    if isinstance(d, str):
        from dateutil.parser import parse, ParserError
        try:
            t = parse(d)
            if t.tzinfo is None and _tzinfo is not None:
                t = t.replace(tzinfo=_tzinfo)
            return t
        except (ParserError, ValueError) as e:
            log.info("Failed to parse string. Attempting to parse as unix time")
            try:
                t = convert_unixtime_datetime(d)
                return t
            except (BaseException, Exception, ParserError) as _err:
                log.warning("Failed to parse unix time. Re-raising original parser error. Unixtime error was: %s %s",
                            type(_err), str(_err))
                raise e
        except ImportError as e:
            msg = "ERROR: Could not import 'parse' from 'dateutil.parser'. Please " \
                  f"make sure 'python-dateutil' is installed. Exception: {type(e)} - {str(e)}"
            
            log.exception(msg)
            raise ImportError(msg)
    if empty(d):
        if fail_empty: raise AttributeError("Error converting datetime. Parameter 'd' was empty!")
        return if_empty
    
    try:
        log.debug("Passed object is not a supported type. Object type: %s || object repr: %s", type(d), repr(d))
        log.debug("Calling convert_datetime with object casted to string: %s", str(d))
        _d = convert_datetime(str(d), fail_empty=True)
        d = _d
    except Exception as e:
        log.info("Converted passed object with str() to try and parse string version, but failed.")
        log.info("Exception thrown from convert_datetime(str(d)) was: %s %s", type(e), str(e))
        d = None   # By setting d to None, it will trigger the ValueError code below.
    
    if not isinstance(d, datetime):
        raise ValueError('Timestamp must be either a datetime object, or an ISO8601 string...')
    return d


parse_datetime = parse_date = convert_datetime


def convert_unixtime_datetime(d: Union[str, int, float, Decimal], if_empty=None, fail_empty=False) -> datetime:
    """Convert a unix timestamp into a :class:`datetime.datetime` object"""
    from dateutil.tz import tzutc
    if empty(d):
        if fail_empty: raise AttributeError("Error converting datetime. Parameter 'd' was empty!")
        return if_empty
    if isinstance(d, datetime):
        return d
    d = int(d)
    # If the timestamp is larger than NOW + 50 years in seconds, then it's probably milliseconds.
    if d > datetime.utcnow().timestamp() + (DECADE * 5):
        t = datetime.utcfromtimestamp(d // 1000)
    else:
        t = datetime.utcfromtimestamp(d)
    
    t = t.replace(tzinfo=tzutc())
    return t


parse_unixtime = parse_epoch = convert_epoch_datetime = convert_unixtime_datetime


def convert_bool_int(d, if_empty=0, fail_empty=False) -> int:
    """Convert a boolean ``d`` into an integer (``0`` for ``False``, ``1`` for ``True``)"""
    if type(d) is int: return 1 if d >= 1 else 0
    if empty(d):
        if fail_empty: raise AttributeError(f"Error converting '{d}' into a boolean. Parameter 'd' was empty!")
        return if_empty
    return 1 if is_true(d) else 0


def convert_int_bool(d, if_empty=False, fail_empty=False) -> bool:
    """Convert an integer ``d`` into a boolean (``0`` for ``False``, ``1`` for ``True``)"""
    if empty(d):
        if fail_empty: raise AttributeError(f"Error converting '{d}' into a boolean. Parameter 'd' was empty!")
        return if_empty
    return is_true(d)


DICT_TYPES = (dict, AttribDictable, Dictable, DictDataClass)

FLOAT_TYPES = (float, Decimal)
INTEGER_TYPES = (int,)
NUMBER_TYPES = FLOAT_TYPES + INTEGER_TYPES

LIST_TYPES = (list, set, tuple)

SIMPLE_TYPES = Union[list, dict, str, float, int]
SIMPLE_TYPES_TUPLE = (list, dict, str, float, int)


def _clean_attrs_matcher(ob):
    try:
        attr.validate(ob)
        return True
        # return clean_dict(attr.asdict(ob))
    except NotAnAttrsClassError:
        return False


_clean_floats = lambda ob, number_str=False, **kwargs: str(ob) if number_str else float(ob)
_clean_ints = lambda ob, number_str=False, **kwargs: str(ob) if number_str else int(ob)


def _clean_strs(ob, **kwargs):
    try:
        return stringify(ob)
    except Exception:
        return str(repr(ob))


def clean_obj(ob: Any, number_str: bool = False, fail=False, fallback: T = None) -> Union[SIMPLE_TYPES, T]:
    """
    Cleans an object by converting it / it's contents into basic, simple, JSON-compatible types.
    
    For example, :class:`.Decimal`'s will become :class:`.float`'s (or :class:`str`'s if ``number_str=True``),
    :class:`bytes` will be decoded into a :class:`str` if possible,
    :param Any ob:              An object to clean - making it safe for use with JSON/YAML etc.
    :param bool number_str:     (Default: ``False``) When set to ``True``, numbers will be converted to strings instead of int/float.
    :param bool fail:           (Default: ``False``) When set to ``True``, will raise the exception thrown by the fallback converter
                                if an error occurs, instead of returning ``fallback``
    :param Any fallback:        (Default: ``None``) The value to return if all matchers/converters fail to handle the object,
                                only used when ``fail=False`` (the default)
    
    :return SIMPLE_TYPES|T res: A clean version of the object for serialisation - or ``fallback`` if something went wrong.
    """
    # if isinstance(ob, FLOAT_TYPES): return str(ob) if number_str else float(ob)
    # if isinstance(ob, INTEGER_TYPES): return str(ob) if number_str else int(ob)
    # if isinstance(ob, NUMBER_TYPES): return str(ob) if number_str else float(ob)
    #
    # if isinstance(ob, (str, bytes)):
    #     try:
    #         return stringify(ob)
    #     except Exception:
    #         return str(repr(ob))
    #
    # if isinstance(ob, DICT_TYPES): return clean_dict(dict(ob))
    # if isinstance(ob, LIST_TYPES): return clean_list(list(ob))
    # if dataclasses.is_dataclass(ob): return dataclasses.asdict(ob)
    
    matched = False
    for matcher, convt in CLEAN_OBJ_VALIDATORS.items():
        try:
            log.debug("Checking matcher: %s - against object: %s", matcher, ob)
            if isinstance(matcher, (list, set)): matcher = tuple(matcher)
            if isinstance(matcher, tuple):
                if not isinstance(ob, matcher): continue
                matched = True
            if not matched and callable(matcher):
                if not matcher(ob): continue
                matched = True
            if not matched:
                if type(ob) is not type(matcher): continue
                matched = True
            log.debug("Matched %s has matched against object. Running converter. Object is: %s", matcher, ob)

            res = convt(ob, number_str=number_str)
            return res
        except ValidatorNotMatched:
            log.info("Matcher %s raised ValidatorNotMatched for object '%s' - continuing.", matcher, ob)
            continue
        except Exception as e:
            log.error("Matcher %s raised %s for object '%s' - continuing. Message was: %s", matcher, type(e), ob, str(e))
            continue
    
    log.warning(
        "All %s matchers failed to match against object '%s' - using fallback converter: %s",
        len(CLEAN_OBJ_VALIDATORS), ob, CLEAN_OBJ_FALLBACK
    )
    try:
        res = CLEAN_OBJ_FALLBACK(ob, number_str=number_str)
        return res
    except Exception as e:
        log.exception("Fallback matcher failed to convert object '%s' ...", ob)
        if fail:
            raise e
        return fallback


def clean_list(ld: list, **kwargs) -> list:
    ld = list(ld)
    nl = []
    for d in ld:
        try:
            x = clean_obj(d, **kwargs)
            nl.append(x)
            # if isinstance(d, (int, float, str)):
            #     nl.append(d)
            #     continue
            # if isinstance(d, LIST_TYPES):
            #     nl.append(clean_list(list(d)))
            #     continue
            # if isinstance(d, LIST_TYPES):
            #     nl.append(clean_list(list(d)))
            #     continue
            # if isinstance(d, DICT_TYPES):
            #     nl.append(clean_dict(dict(d)))
            #     continue
            # nl.append(str(d))
        except Exception:
            log.exception("Error while cleaning list item: %s", d)
    return nl


def clean_dict(data: dict, **kwargs) -> dict:
    data = dict(data)
    cleaned = {}
    for k, v in data.items():
        try:
            n = clean_obj(v, **kwargs)
            cleaned[k] = n
            # if isinstance(v, (dict, AttribDictable)):
            #     n = clean_dict(dict(v))
            #     cleaned[k] = n
            #     continue
            # if isinstance(v, list):
            #     n = clean_list(list(v))
            #     cleaned[k] = n
            #     continue
            # if isinstance(v, (int, float, str)):
            #     cleaned[k] = v
            #     continue
            # cleaned[k] = str(v)
        except Exception:
            log.exception("Error while cleaning dict item: %s = %s", k, v)
    
    return cleaned


CLEAN_OBJ_VALIDATORS = {
    FLOAT_TYPES:          _clean_floats,
    INTEGER_TYPES:        _clean_ints,
    NUMBER_TYPES:         _clean_floats,
    (str, bytes):         _clean_strs,
    DICT_TYPES:           clean_dict,
    LIST_TYPES:           clean_list,
    _clean_attrs_matcher:     lambda ob, **kwargs: clean_obj(attr.asdict(ob), **kwargs),
    dataclasses.is_dataclass: lambda ob, **kwargs: clean_obj(dataclasses.asdict(ob), **kwargs),
}

CLEAN_OBJ_FALLBACK = lambda ob, **kwargs: str(ob)

__all__ = [
    'convert_datetime', 'convert_unixtime_datetime', 'convert_bool_int', 'convert_int_bool',
    'parse_date', 'parse_datetime', 'parse_epoch', 'parse_unixtime', 'convert_epoch_datetime',
    'DICT_TYPES', 'FLOAT_TYPES', 'INTEGER_TYPES', 'NUMBER_TYPES', 'LIST_TYPES',
    'clean_obj', 'clean_list', 'clean_dict', 'CLEAN_OBJ_FALLBACK', 'CLEAN_OBJ_VALIDATORS',
    'MINUTE', 'HOUR', 'DAY', 'MONTH', 'YEAR', 'DECADE',
    
]




"""
Class Method / Function decorators

**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
import asyncio
import functools
import logging
from enum import Enum
from time import sleep
from typing import Any, Union, List

from privex.helpers.cache import cached, async_adapter_get
from privex.helpers.common import empty, is_true
from privex.helpers.asyncx import await_if_needed

DEF_RETRY_MSG = "Exception while running '%s', will retry %d more times."
DEF_FAIL_MSG = "Giving up after attempting to retry function '%s' %d times."

log = logging.getLogger(__name__)

try:
    from privex.helpers.asyncx import async_sync
except ImportError:
    log.debug('privex.helpers __init__ failed to import "asyncx", not loading async helpers')
    pass


def retry_on_err(max_retries: int = 3, delay: Union[int, float] = 3, **retry_conf):
    """
    Decorates a function or class method, wraps the function/method with a try/catch block, and will automatically
    re-run the function with the same arguments up to `max_retries` time after any exception is raised, with a
    ``delay`` second delay between re-tries.

    If it still throws an exception after ``max_retries`` retries, it will log the exception details with ``fail_msg``,
    and then re-raise it.

    Usage (retry up to 5 times, 1 second between retries, stop immediately if IOError is detected)::

        >>> @retry_on_err(5, 1, fail_on=[IOError])
        ... def my_func(self, some=None, args=None):
        ...     if some == 'io': raise IOError()
        ...      raise FileExistsError()

    This will be re-ran 5 times, 1 second apart after each exception is raised, before giving up::

        >>> my_func()

    Where-as this one will immediately re-raise the caught IOError on the first attempt, as it's passed in ``fail_on``::

        >>> my_func('io')


    .. Attention:: For safety reasons, by default ``max_ignore`` is set to ``100``. This means after 100 retries where an
                       exception was ignored, the decorator will give up and raise the last exception.
                       
                       This is to prevent the risk of infinite loops hanging your application. If you are 100% certain that the
                       function you've wrapped, and/or the exceptions passed in ``ignore`` cannot cause an infinite retry loop, then
                       you can pass ``max_ignore=False`` to the decorator to disable failure after ``max_ignore`` ignored exceptions.
    


    :param int max_retries:  Maximum total retry attempts before giving up
    :param float delay:      Amount of time in seconds to sleep before re-trying the wrapped function
    :param retry_conf:       Less frequently used arguments, pass in as keyword args (see below)

    :key list fail_on:  A list() of Exception types that should result in immediate failure (don't retry, raise)
    
    :key list ignore:   A list() of Exception types that should be ignored (will retry, but without incrementing the failure counter)
    
    :key int|bool max_ignore: (Default: ``100``) If an exception is raised while retrying, and more than this
             many exceptions (listed in ``ignore``) have been ignored during retry attempts, then give up
             and raise the last exception.
             
             This feature is designed to prevent "ignored" exceptions causing an infinite retry loop. By
             default ``max_ignore`` is set to ``100``, but you can increase/decrease this as needed.
             
             You can also set it to ``False`` to disable raising when too many exceptions are ignored - however, it's
             strongly not recommended to disable ``max_ignore``, especially if you have ``instance_match=True``,
             as it could cause an infinite retry loop which hangs your application.
      
    
    :key bool instance_match: (Default: ``False``) If this is set to ``True``, then the exception type comparisons for ``fail_on``
            and ``ignore`` will compare using ``isinstance(e, x)`` instead of ``type(e) is x``.
            
            If this is enabled, then exceptions listed in ``fail_on`` and ``ignore`` will also **match sub-classes** of
            the listed exceptions, instead of exact matches.

    :key str retry_msg: Override the log message used for retry attempts. First message param %s is func name,
                        second message param %d is retry attempts remaining

    :key str fail_msg: Override the log message used after all retry attempts are exhausted. First message param %s
                       is func name, and second param %d is amount of times retried.

    """
    retry_msg: str = retry_conf.get('retry_msg', DEF_RETRY_MSG)
    fail_msg: str = retry_conf.get('fail_msg', DEF_FAIL_MSG)
    instance_match: bool = is_true(retry_conf.get('instance_match', False))
    fail_on: List[type] = list(retry_conf.get('fail_on', []))
    ignore_ex: List[type] = list(retry_conf.get('ignore', []))
    max_ignore: Union[bool, int] = retry_conf.get('max_ignore', 100)

    def _decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            retries = int(kwargs.pop('retry_attempts', 0))
            ignore_count = int(kwargs.pop('ignore_count', 0))
            
            try:
                return f(*args, **kwargs)
            except Exception as e:
                _fail = isinstance(e, tuple(fail_on)) if instance_match else type(e) in fail_on
                if _fail:
                    log.warning('Giving up. Re-raising exception %s (as requested by `fail_on` arg)', type(e))
                    raise e
                
                if max_ignore is not False and ignore_count > max_ignore:
                    log.warning('Giving up. Ignored too many exceptions (max_ignore: %d, ignore_count: %d). '
                                'Re-raising exception %s.', max_ignore, ignore_count, type(e))
                    raise e
                
                if retries < max_retries:
                    log.info('%s - %s', type(e), str(e))
                    log.info(retry_msg, f.__name__, max_retries - retries)
                    sleep(delay)
                    
                    # If 'instance_match' is enabled, we check if the exception was an instance of any of the passed exception types,
                    # otherwise we use exact exception type comparison against the list.
                    # _ignore is True if we should ignore this exception (don't increment retries), or False if we should increment.
                    _ignore = isinstance(e, tuple(ignore_ex)) if instance_match else type(e) in ignore_ex
                    if _ignore:
                        log.debug(
                            " >> (?) Ignoring exception '%s' as exception is in 'ignore' list. Ignore Count: %d // "
                            "Max Ignores: %d // Instance Match: %s", type(e), ignore_count, max_ignore, instance_match
                        )
                    kwargs['retry_attempts'] = retries if _ignore else retries + 1
                    kwargs['ignore_count'] = ignore_count + 1 if _ignore else ignore_count
                    
                    return wrapper(*args, **kwargs)
                log.exception(fail_msg, f.__name__, max_retries)
                raise e
        return wrapper
    return _decorator


def async_retry(max_retries: int = 3, delay: Union[int, float] = 3, **retry_conf):
    """
    AsyncIO coroutine compatible version of :func:`.retry_on_err` - for painless automatic retry-on-exception for async code.
    
    Decorates an AsyncIO coroutine (``async def``) function or class method, wraps the function/method with a try/catch block, and
    will automatically re-run the function with the same arguments up to `max_retries` time after any exception is raised, with a
    ``delay`` second delay between re-tries.

    If it still throws an exception after ``max_retries`` retries, it will log the exception details with ``fail_msg``,
    and then re-raise it.
    
    Usage (retry up to 5 times, 1 second between retries, stop immediately if IOError is detected)::
    
        >>> from privex.helpers import async_retry
        >>>
        >>> @async_retry(5, 1, fail_on=[IOError])
        ... async def my_func(some=None, args=None):
        ...     if some == 'io': raise IOError()
        ...     raise FileExistsError()
        ...
    
    This will be re-ran 5 times, 1 second apart after each exception is raised, before giving up::

        >>> await my_func()

    Where-as this one will immediately re-raise the caught IOError on the first attempt, as it's passed in ``fail_on``::

        >>> await my_func('io')
    
    We can also use ``ignore_on`` to "ignore" certain exceptions. Ignored exceptions cause the function to be retried with a delay, as
    normal, but without incrementing the total retries counter.
    
        >>> from privex.helpers import async_retry
        >>> import random
        >>>
        >>> @async_retry(5, 1, fail_on=[IOError], ignore=[ConnectionResetError])
        ... async def my_func(some=None, args=None):
        ...     if random.randint(1,10) > 7: raise ConnectionResetError()
        ...     if some == 'io': raise IOError()
        ...     raise FileExistsError()
        ...
    
    To show this at work, we've enabled debug logging for you to see::
        
        >>> await my_func()
        [INFO]    <class 'ConnectionResetError'> -
        [INFO]    Exception while running 'my_func', will retry 5 more times.
        [DEBUG]   >> (?) Ignoring exception '<class 'ConnectionResetError'>' as exception is in 'ignore' list.
                  Ignore Count: 0 // Max Ignores: 100 // Instance Match: False
        
        [INFO]    <class 'FileExistsError'> -
        [INFO]    Exception while running 'my_func', will retry 5 more times.
        
        [INFO]    <class 'ConnectionResetError'> -
        [INFO]    Exception while running 'my_func', will retry 4 more times.
        [DEBUG]   >> (?) Ignoring exception '<class 'ConnectionResetError'>' as exception is in 'ignore' list.
                  Ignore Count: 1 // Max Ignores: 100 // Instance Match: False
        
        [INFO]    <class 'FileExistsError'> -
        [INFO]    Exception while running 'my_func', will retry 4 more times.
    
    As you can see above, when an ignored exception (``ConnectionResetError``) occurs, the remaining retry attempts doesn't go down.
    Instead, only the "Ignore Count" goes up.
    
    .. Attention:: For safety reasons, by default ``max_ignore`` is set to ``100``. This means after 100 retries where an
                   exception was ignored, the decorator will give up and raise the last exception.
                   
                   This is to prevent the risk of infinite loops hanging your application. If you are 100% certain that the
                   function you've wrapped, and/or the exceptions passed in ``ignore`` cannot cause an infinite retry loop, then
                   you can pass ``max_ignore=False`` to the decorator to disable failure after ``max_ignore`` ignored exceptions.
    
    
    :param int max_retries:  Maximum total retry attempts before giving up
    :param float delay:      Amount of time in seconds to sleep before re-trying the wrapped function
    :param retry_conf:       Less frequently used arguments, pass in as keyword args (see below)
    
    :key list fail_on:  A list() of Exception types that should result in immediate failure (don't retry, raise)
    
    :key list ignore:   A list() of Exception types that should be ignored (will retry, but without incrementing the failure counter)

    :key int|bool max_ignore: (Default: ``100``) If an exception is raised while retrying, and more than this
             many exceptions (listed in ``ignore``) have been ignored during retry attempts, then give up
             and raise the last exception.
             
             This feature is designed to prevent "ignored" exceptions causing an infinite retry loop. By
             default ``max_ignore`` is set to ``100``, but you can increase/decrease this as needed.
             
             You can also set it to ``False`` to disable raising when too many exceptions are ignored - however, it's
             strongly not recommended to disable ``max_ignore``, especially if you have ``instance_match=True``,
             as it could cause an infinite retry loop which hangs your application.
      
    
    :key bool instance_match: (Default: ``False``) If this is set to ``True``, then the exception type comparisons for ``fail_on``
            and ``ignore`` will compare using ``isinstance(e, x)`` instead of ``type(e) is x``.
            
            If this is enabled, then exceptions listed in ``fail_on`` and ``ignore`` will also **match sub-classes** of
            the listed exceptions, instead of exact matches.

    :key str retry_msg: Override the log message used for retry attempts. First message param %s is func name,
                        second message param %d is retry attempts remaining

    :key str fail_msg: Override the log message used after all retry attempts are exhausted. First message param %s
                       is func name, and second param %d is amount of times retried.
    
    """
    retry_msg: str = retry_conf.get('retry_msg', DEF_RETRY_MSG)
    fail_msg: str = retry_conf.get('fail_msg', DEF_FAIL_MSG)
    instance_match: bool = is_true(retry_conf.get('instance_match', False))
    fail_on: List[type] = list(retry_conf.get('fail_on', []))
    ignore_ex: List[type] = list(retry_conf.get('ignore', []))
    max_ignore: Union[bool, int] = retry_conf.get('max_ignore', 100)
    
    def _decorator(f):
        @functools.wraps(f)
        async def wrapper(*args, **kwargs):
            retries = int(kwargs.pop('retry_attempts', 0))
            ignore_count = int(kwargs.pop('ignore_count', 0))
            
            try:
                return await f(*args, **kwargs)
            except Exception as e:
                # If instance_match is enabled, check exception type using isinstance, otherwise use exact type matches
                _fail = isinstance(e, tuple(fail_on)) if instance_match else type(e) in fail_on
                if _fail:
                    log.warning('Giving up. Re-raising exception %s (as requested by `fail_on` arg)', type(e))
                    raise e
                
                if max_ignore is not False and ignore_count > max_ignore:
                    log.warning('Giving up. Ignored too many exceptions (max_ignore: %d, ignore_count: %d). '
                                'Re-raising exception %s.', max_ignore, ignore_count, type(e))
                    raise e
                
                if retries < max_retries:
                    log.info('%s - %s', type(e), str(e))
                    log.info(retry_msg, f.__name__, max_retries - retries)
                    await asyncio.sleep(delay)
                    
                    # If 'instance_match' is enabled, we check if the exception was an instance of any of the passed exception types,
                    # otherwise we use exact exception type comparison against the list.
                    # _ignore is True if we should ignore this exception (don't increment retries), or False if we should increment.
                    _ignore = isinstance(e, tuple(ignore_ex)) if instance_match else type(e) in ignore_ex
                    if _ignore:
                        log.debug(
                            " >> (?) Ignoring exception '%s' as exception is in 'ignore' list. Ignore Count: %d // "
                            "Max Ignores: %d // Instance Match: %s", type(e), ignore_count, max_ignore, instance_match
                        )
                    kwargs['retry_attempts'] = retries if _ignore else retries + 1
                    kwargs['ignore_count'] = ignore_count + 1 if _ignore else ignore_count
                    
                    return await wrapper(*args, **kwargs)
                log.exception(fail_msg, f.__name__, max_retries)
                raise e
        
        return wrapper
    
    return _decorator


class FormatOpt(Enum):
    """
    This enum represents various options available for :py:func:`.r_cache` 's ``format_opt`` parameter.

    To avoid bloating the PyDoc for ``r_cache`` too much, descriptions for each formatting option is available as a
    short PyDoc comment under each enum option.

    Usage:

    >>> @r_cache('mykey', format_args=[0, 'x'], format_opt=FormatOpt.POS_AUTO)

    """
    POS_AUTO = 'force_pos'
    """
    First attempt to format using ``*args`` whitelisted in ``format_args``, if that causes a KeyError/IndexError,
    then pass kwarg values in the order they're listed in ``format_args``
    (only includes kwarg names listed in ``format_args``)

    # def func(x, y)
    func('a', 'b')      # assuming 0 and 1 are in format_args, then it would use .format('a', 'b')
    func(y='b', x='a')  # assuming format_args = ``['x','y']``, then it would use .format('a', 'b')
    """

    POS_ONLY = 'pos_only'
    """Only use positional args for formatting the cache key, kwargs will be ignored completely."""

    KWARG_ONLY = 'kwarg'
    """Only use kwargs for formatting the cache key - requires named format placeholders, i.e. ``mykey:{x}``"""

    MIX = 'mix'
    """
    Use both ``*args`` and ``**kwargs`` to format the cache_key (assuming mixed placeholders e.g. ``mykey:{}:{y}``
    """


FO = FormatOpt


def _format_key(args, kwargs, cache_key: str, whitelist: bool = True, fmt_opt: FO = FO.POS_AUTO, fmt_args: list = None):
    """Internal function used by :func:`.r_cache` and :func:`.r_cache_async` for formatting a cache key e.g. ``pvx:{}:{}``"""
    pos_args = args
    kw_args = kwargs
    if whitelist:
        # Whitelisted positional arguments
        pos_args = [args[i] for i in fmt_args if type(i) is int and len(args) > i]
        # Whitelisted keyword args, as a dict
        kw_args = {i: kwargs[i] for i in fmt_args if type(i) is str and i in kwargs}
    
    if fmt_opt == FormatOpt.POS_AUTO:
        log.debug('Format: POS_AUTO - Formatting with *args, fallback on err to positional **kwargs values')
        try:
            log.debug('Attempting to format with args: %s', pos_args)
            rk = cache_key.format(*pos_args)
        except (KeyError, IndexError):
            pos_kwargs = [v for _, v in kw_args.items()]
            log.debug('Failed to format with pos args, now trying positional kwargs: %s', pos_kwargs)
            rk = cache_key.format(*pos_kwargs)
        return rk
    
    if fmt_opt == FormatOpt.KWARG_ONLY:  # Only attempt to format cache_key using kwargs
        log.debug('Format: KWARG_ONLY - Formatting using only **kwargs')
        return cache_key.format(**kw_args)
    
    if fmt_opt == FormatOpt.POS_ONLY:  # Only attempt to format cache_key using positional args
        log.debug('Format: POS_ONLY - Formatting using only *args')
        return cache_key.format(*pos_args)
    
    if fmt_opt == FormatOpt.MIX:  # Format cache_key with both positional and kwargs as-is
        log.debug('Format: MIX - Formatting using passthru *args and **kwargs')
        return cache_key.format(*pos_args, **kw_args)


def r_cache_async(cache_key: Union[str, callable], cache_time=300, format_args: list = None, format_opt: FO = FO.POS_AUTO, **opts) -> Any:
    """
    Async function/method compatible version of :func:`.r_cache` - see docs for :func:`.r_cache`
    
    You can bypass caching by passing ``r_cache=False`` to the wrapped function.
    
    Basic usage::
    
        >>> from privex.helpers import r_cache_async
        >>> @r_cache_async('my_cache_key')
        >>> async def some_func(some: int, args: int = 2):
        ...     return some + args
        >>> await some_func(5, 10)
        15
        
        >>> # If we await some_func a second time, we'll get '15' again because it was cached.
        >>> await some_func(2, 3)
        15
    
    Async ``cache_key`` generation (you can also use normal synchronous functions/lambdas)::
    
        >>> from privex.helpers import r_cache_async
        >>>
        >>> async def make_key(name, title):
        ...     return f"mycache:{name}"
        ...
        >>> @r_cache_async(make_key)
        ... async def who(name, title):
        ...     return "Their name is {title} {name}"
        ...
        
    :param FormatOpt format_opt: (default: :py:attr:`.FormatOpt.POS_AUTO`) "Format option" - how should args/kwargs be
                                 used when filling placeholders in the ``cache_key`` (see comments on FormatOption)
    :param list format_args: A list of positional arguments numbers (e.g. ``[0, 1, 2]``) and/or kwargs
                             ``['x', 'y', 'z']`` that should be used to format the `cache_key`
    :param str cache_key: The cache key to store the cached data into, e.g. `mydata`
    :param int cache_time: The amount of time in seconds to cache the result for (default: 300 seconds)
    :keyword bool whitelist: (default: ``True``) If True, only use specified arg positions / kwarg keys when formatting
                             ``cache_key`` placeholders. Otherwise, trust whatever args/kwargs were passed to the func.
    :return Any res: The return result, either from the wrapped function, or from the cache.
    """
    fmt_args = [] if not format_args else format_args
    # Using normal 'cached' often results in "event loop already running" errors due to the synchronous async wrapper
    # in CacheWrapper. So to be safe, we get the adapter directly to avoid issues.
    cache_adapter = async_adapter_get()
    whitelist = opts.get('whitelist', True)
    
    def _decorator(f):
        @functools.wraps(f)
        async def wrapper(*args, **kwargs):
            
            # Extract r_cache and r_cache_key from the wrapped function's kwargs if they're specified,
            # then remove them from the kwargs so they don't interfere with the wrapped function.
            enable_cache, rk = kwargs.get('r_cache', True), kwargs.get('r_cache_key', cache_key)
            if 'r_cache' in kwargs: del kwargs['r_cache']
            if 'r_cache_key' in kwargs: del kwargs['r_cache_key']

            if not isinstance(rk, str):
                rk = await await_if_needed(rk, *args, **kwargs)
            elif not empty(fmt_args, itr=True) or not whitelist:
                # If the cache key contains a format placeholder, e.g. {somevar} - then attempt to replace the
                # placeholders using the function's kwargs
                log.debug('Format_args not empty (or whitelist=False), formatting cache_key "%s"', cache_key)
                rk = _format_key(args, kwargs, cache_key=cache_key, whitelist=whitelist, fmt_opt=format_opt, fmt_args=format_args)
            # To ensure no event loop / thread cache instance conflicts, we use the cache adapter as a context manager, which
            # is supposed to disconnect + destroy the connection library instance, and re-create it in the current loop/thread.
            async with cache_adapter as r:
                # If using an async cache adapter, r.get might be async...
                log.debug('Trying to load "%s" from cache', rk)
                data = await await_if_needed(r.get, rk)
                
                if empty(data) or not enable_cache:
                    log.debug('Not found in cache, or "r_cache" set to false. Calling wrapped async function.')
                    data = await await_if_needed(f, *args, **kwargs)
    
                    # If using an async cache adapter, r.get might be async...
                    await await_if_needed(r.set, rk, data, timeout=cache_time)
            
            return data
        
        return wrapper
    
    return _decorator


def r_cache(cache_key: Union[str, callable], cache_time=300, format_args: list = None,
            format_opt: FO = FO.POS_AUTO, **opts) -> Any:
    """
    This is a decorator which caches the result of the wrapped function with the global cache adapter from
    :py:mod:`privex.helpers.cache` using the key ``cache_key`` and with an expiry of ``cache_time`` seconds.

    Future calls to the wrapped function would then load the data from cache until the cache expires, upon which it
    will re-run the original code and re-cache it.

    To bypass the cache, pass kwarg ``r_cache=False`` to the wrapped function. To override the cache key on demand,
    pass ``r_cache_key='mykey'`` to the wrapped function.

    **Example usage**:

        >>> from privex.helpers import r_cache
        >>>
        >>> @r_cache('mydata', cache_time=600)
        ... def my_func(*args, **kwargs):
        ...     time.sleep(60)
        ...     return "done"

        This will run the function and take 60 seconds to return while it sleeps

        >>> my_func()
        done

        This will run instantly because "done" is now cached for 600 seconds

        >>> my_func()
        done

        This will take another 60 seconds to run because ``r_cache`` is set to `False` (disables the cache)

        >>> my_func(r_cache=False)
        done

    **Using a dynamic cache_key**:

        **Simplest and most reliable - pass ``r_cache_key`` as an additional kwarg**

        If you don't mind passing an additional kwarg to your function, then the most reliable method is to override
        the cache key by passing ``r_cache_key`` to your wrapped function.

        Don't worry, we remove both ``r_cache`` and ``r_cache_key`` from the kwargs that actually hit your function.

        >>> my_func(r_cache_key='somekey')    # Use the cache key 'somekey' when caching data for this function

        **Option 2. Pass a callable which takes the same arguments as the wrapped function**

        In the example below, ``who`` takes two arguments: ``name`` and ``title`` - we then pass the function
        ``make_key`` which takes the same arguments - ``r_cache`` will detect that the cache key is a function
        and call it with the same ``(*args, **kwargs)`` passed to the wrapped function.

        >>> from privex.helpers import r_cache
        >>>
        >>> def make_key(name, title):
        ...     return f"mycache:{name}"
        ...
        >>> @r_cache(make_key)
        ... def who(name, title):
        ...     return "Their name is {title} {name}"
        ...

        We can also obtain the same effect with a ``lambda`` callable defined directly inside of the cache_key.

        >>> @r_cache(lambda name,title: f"mycache:{name}")
        ... def who(name, title):
        ...     return "Their name is {title} {name}"

        **Option 3. Can be finnicky - using ``format_args`` to integrate with existing code**

        If you can't change how your existing function/method is called, then you can use the ``format_args`` feature.

        **NOTE:** Unless you're forcing the usage of kwargs with a function/method, it's strongly recommended that you
        keep ``force_pos`` enabled, and specify both the positional argument ID, and the kwarg name.

        Basic Example:

        >>> from privex.helpers import r_cache
        >>> import time
        >>>
        >>> @r_cache('some_cache:{}:{}', cache_time=600, format_args=[0, 1, 'x', 'y'])
        ... def some_func(x=1, y=2):
        ...     time.sleep(5)
        ...     return 'x + y = {}'.format(x + y)
        >>>

        Using positional arguments, we can see from the debug log that it's formatting the ``{}:{}`` in the key
        with ``x:y``

        >>> some_func(1, 2)
        2019-08-21 06:58:29,823 lg  DEBUG    Trying to load "some_cache:1:2" from cache
        2019-08-21 06:58:29,826 lg  DEBUG    Not found in cache, or "r_cache" set to false. Calling wrapped function.
        'x + y = 3'
        >>> some_func(2, 3)
        2019-08-21 06:58:34,831 lg  DEBUG    Trying to load "some_cache:2:3" from cache
        2019-08-21 06:58:34,832 lg  DEBUG    Not found in cache, or "r_cache" set to false. Calling wrapped function.
        'x + y = 5'

        When we passed ``(1, 2)`` and ``(2, 3)`` it had to re-run the function for each. But once we re-call it for
        the previously ran ``(1, 2)`` - it's able to retrieve the cached result just for those args.

        >>> some_func(1, 2)
        2019-08-21 06:58:41,752 lg  DEBUG    Trying to load "some_cache:1:2" from cache
        'x + y = 3'

        Be warned that the default format option ``POS_AUTO`` will make kwargs' values be specified in the same order as
        they were listed in ``format_args``

        >>> some_func(y=1, x=2)   # ``format_args`` has the kwargs in the order ``['x', 'y']`` thus ``.format(x,y)``
        2019-08-21 06:58:58,611 lg  DEBUG    Trying to load "some_cache:2:1" from cache
        2019-08-21 06:58:58,611 lg  DEBUG    Not found in cache, or "r_cache" set to false. Calling wrapped function.
        'x + y = 3'


    :param FormatOpt format_opt: (default: :py:attr:`.FormatOpt.POS_AUTO`) "Format option" - how should args/kwargs be
                                 used when filling placeholders in the ``cache_key`` (see comments on FormatOption)
    :param list format_args: A list of positional arguments numbers (e.g. ``[0, 1, 2]``) and/or kwargs
                             ``['x', 'y', 'z']`` that should be used to format the `cache_key`
    :param str cache_key: The cache key to store the cached data into, e.g. `mydata`
    :param int cache_time: The amount of time in seconds to cache the result for (default: 300 seconds)
    :keyword bool whitelist: (default: ``True``) If True, only use specified arg positions / kwarg keys when formatting
                             ``cache_key`` placeholders. Otherwise, trust whatever args/kwargs were passed to the func.
    :return Any res: The return result, either from the wrapped function, or from the cache.
    """
    fmt_args = [] if not format_args else format_args
    r = cached
    whitelist = opts.get('whitelist', True)

    def _decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):

            # Extract r_cache and r_cache_key from the wrapped function's kwargs if they're specified,
            # then remove them from the kwargs so they don't interfere with the wrapped function.
            enable_cache, rk = kwargs.get('r_cache', True), kwargs.get('r_cache_key', cache_key)
            if 'r_cache' in kwargs: del kwargs['r_cache']
            if 'r_cache_key' in kwargs: del kwargs['r_cache_key']

            if callable(rk):
                rk = rk(*args, **kwargs)
            elif not empty(fmt_args, itr=True) or not whitelist:
                # If the cache key contains a format placeholder, e.g. {somevar} - then attempt to replace the
                # placeholders using the function's kwargs
                log.debug('Format_args not empty (or whitelist=False), formatting cache_key "%s"', cache_key)
                rk = _format_key(args, kwargs, cache_key=cache_key, whitelist=whitelist, fmt_opt=format_opt, fmt_args=format_args)
            log.debug('Trying to load "%s" from cache', rk)
            data = r.get(rk)

            if empty(data) or not enable_cache:
                log.debug('Not found in cache, or "r_cache" set to false. Calling wrapped function.')
                data = f(*args, **kwargs)
                r.set(rk, data, timeout=cache_time)

            return data

        return wrapper

    return _decorator


def mock_decorator(*dec_args, **dec_kwargs):
    """
    This decorator is a pass-through decorator which does nothing other than be a decorator.
    
    It's designed to be used with the :class:`privex.helpers.common.Mocker` class when mocking classes/modules,
    allowing you to add fake decorators to the mock class/method which do nothing, other than act like a decorator
    without breaking your functions/methods.
    """
    def _decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            return f(*args, **kwargs)
        return wrapper
    return _decorator

